import dexy.exceptions
import hashlib
import inspect
import json
import logging
import os
import platform
import posixpath
import re
import shutil
import tempfile
import time
import yaml

is_windows = platform.system() in ('Windows',)

def copy_or_link(data, destination, use_links=False, read_only_links=True):
    """
    Copies or makes a hard link. Will copy if on windows or if use_links is False.
    """
    if is_windows or not use_links:
        data.output_to_file(destination)
    else:
        os.link(data.storage.data_file(), destination)

defaults = {
    'artifacts_dir' : '.dexy',
    'config_file' : 'dexy.conf',
    'configs' : '',
    'debug' : False,
    'directory' : ".",
    'dont_use_cache' : False,
    'dry_run' : False,
    'encoding' : 'utf-8',
    'exclude' : '.git, .svn, tmp, cache, .trash, .ipynb_checkpoints',
    'exclude_also' : '',
    'full' : False,
    'globals' : '',
    'hashfunction' : 'md5',
    'ignore_nonzero_exit' : False,
    'include' : '',
    'log_dir' : '.dexy',
    'log_file' : 'dexy.log',
    'log_format' : "%(name)s - %(levelname)s - %(message)s",
    'log_level' : "INFO",
    'output_root' : '.',
    'parsers' : "dexy-env.json dexy.txt dexy.yaml",
    'pickle' : 'c',
    'plugins': 'dexyplugins.py dexyplugin.py dexyplugins.yaml dexyplugin.yaml',
    'profile' : False,
    'recurse' : True,
    'reports' : '',
    'safety_filename' : '.dexy-generated',
    'siblings' : False,
    'silent' : False,
    'strace' : False,
    'target' : False,
    'timing' : True,
    'uselocals' : False,
    'writeanywhere' : False
}

log_levels = {
    'DEBUG' : logging.DEBUG,
    'INFO' : logging.INFO,
    'WARN' : logging.WARN
}

def transition(obj, new_state):
    """
    Attempts to transition this object to the new state, if the transition
    from current state to new state is valid as per state_transitions list.
    """
    attempted_transition = (obj.state, new_state) 
    if not attempted_transition in obj.__class__.state_transitions:
        msg = "%s -> %s"
        raise dexy.exceptions.UnexpectedState(msg % attempted_transition)

    if not hasattr(obj, 'time_entered_current_state'):
        obj.time_entered_current_state = None
        obj.state_history = []
  
    if obj.time_entered_current_state:
        transition_time = time.time()

        time_in_prev_state = transition_time - obj.time_entered_current_state
        obj.state_history.append((obj.state, time_in_prev_state))

    obj.time_entered_current_state = time.time()
    obj.state = new_state

def logging_log_level(log_level):
    try:
        return log_levels[log_level.upper()]
    except KeyError:
        msg = "'%s' is not a valid log level, check python logging module docs"
        raise dexy.exceptions.UserFeedback(msg % log_level)

def md5_hash(text):
    return hashlib.md5(text.encode('utf-8')).hexdigest()

def dict_from_string(text):
    """
    Creates a dict from string like "key1=value1,k2=v2"
    """
    d = {}
    if text > 0:
        for pair in text.split(","):
            x, y = pair.split("=")
            d[x] = y
    return d

def file_exists(filepath, debug=False):
    if debug:
        print(("calling file_exists on %s" % filepath))
        frame = inspect.currentframe()
        for f in inspect.getouterframes(frame):
            print(("   ", f[1], f[2], f[3]))
        del frame
        del f
    return os.path.exists(filepath)

def iter_paths(path):
    """
    Iterate over all subpaths of path starting with root path.
    """
    path_elements = split_path(path)

    if path.startswith(os.sep):
        start = os.sep
    else:
        start = None

    for i in range(1, len(path_elements)+1):
        if start:
            yield os.path.join(start, *path_elements[0:i])
        else:
            yield os.path.join(*path_elements[0:i])

def reverse_iter_paths(path):
    """
    Iterate over all subpaths of path starting with path, ending with root path.
    """
    path_elements = split_path(path)
    for i in range(len(path_elements), 0, -1):
        yield os.path.join(*path_elements[0:i])
    yield "/"

def split_path(path):
    # TODO test that paths are normed and don't include empty or '.' components.
    tail = True
    path_elements = []
    body = path
    while tail:
        body, tail = os.path.split(body)
        if tail:
            path_elements.append(tail)
        elif path.startswith("/"):
            path_elements.append(tail)
            
    path_elements.reverse()
    return path_elements

class tempdir(object):
    def make_temp_dir(self):
        self.tempdir = tempfile.mkdtemp()
        self.location = os.path.abspath(os.curdir)
        os.chdir(self.tempdir)

    def remove_temp_dir(self):
        os.chdir(self.location)
        try:
            shutil.rmtree(self.tempdir)
        except Exception as e:
            print(e)
            print(("was not able to remove tempdir '%s'" % self.tempdir))

    def __enter__(self):
        self.make_temp_dir()

    def __exit__(self, type, value, traceback):
        if not isinstance(value, Exception):
            self.remove_temp_dir()

class chdir(object):
    def __init__(self, newdir):
        self.newdir = newdir
        self.location = os.path.abspath(os.curdir)

    def __enter__(self):
        os.chdir(self.newdir)

    def __exit__(self, type, value, traceback):
        os.chdir(self.location)

def value_for_hyphenated_or_underscored_arg(arg_dict, arg_name_hyphen, default=None):
    if not "-" in arg_name_hyphen and "_" in arg_name_hyphen:
        raise dexy.exceptions.InternalDexyProblem("arg_name_hyphen %s has underscores!" % arg_name_hyphen)

    arg_name_underscore = arg_name_hyphen.replace("-", "_")

    arg_value = arg_dict.get(arg_name_hyphen)

    if arg_value is None:
        arg_value = arg_dict.get(arg_name_underscore)

    if arg_value is None:
        return default
    else:
        return arg_value

def s(text):
    return re.sub('\s+', ' ', text)

def getdoc(element, firstline=True):
    docstring = inspect.getdoc(element)
    if docstring and firstline:
        docstring = docstring.splitlines()[0]

    if not docstring:
        docstring = ''

    return docstring

def os_to_posix(path):
    return posixpath.join(*os.path.split(path))

def parse_json(input_text):
    try:
        return json.loads(input_text)
    except ValueError as e:
        msg = inspect.cleandoc("""Was unable to parse the JSON you supplied.
        Here is information from the JSON parser:""")
        msg += "\n"
        msg += str(e)
        raise dexy.exceptions.UserFeedback(msg)

def parse_json_from_file(f):
    try:
        return json.load(f)
    except ValueError as e:
        msg = inspect.cleandoc("""Was unable to parse the JSON you supplied.
        Here is information from the JSON parser:""")
        msg += "\n"
        msg += str(e)
        raise dexy.exceptions.UserFeedback(msg)

def parse_yaml(input_text):
    """
    Parse a single YAML document.
    """
    try:
        return yaml.safe_load(input_text)
    except (yaml.scanner.ScannerError, yaml.parser.ParserError) as e:
        if "found character '\\t'" in str(e):
            msg = "You appear to have hard tabs in your yaml, this is not supported. Please change to using soft tabs instead (your text editor should have this option)."
            raise dexy.exceptions.UserFeedback(msg)
        else:
            msg = inspect.cleandoc("""Was unable to parse the YAML you supplied.
            Here is information from the YAML parser:""")
            msg += "\n"
            msg += str(e)
            raise dexy.exceptions.UserFeedback(msg)

def parse_yamls(input_text):
    """
    Parse YAML content that may include more than 1 document.
    """
    try:
        return yaml.safe_load_all(input_text)
    except (yaml.scanner.ScannerError, yaml.parser.ParserError) as e:
        msg = inspect.cleandoc("""Was unable to parse the YAML you supplied.
        Here is information from the YAML parser:""")
        msg += "\n"
        msg += str(e)
        raise dexy.exceptions.UserFeedback(msg)

def printable_for_char(c):
    if ord(c) >= ord('!'):
        return c
    elif ord(c) == 32:
        return "<space>"
    else:
        return ""

def char_diff(str1, str2):
    """
    Returns a char-by-char diff of two strings, highlighting differences.
    """
    msg = ""
    for i, c1 in enumerate(str1):
        if len(str2) > i:
            c2 = str2[i]

            if c1 == c2:
                flag = ""
            else:
                flag = " <---"

            p_c1 = printable_for_char(c1)
            p_c2 = printable_for_char(c2)

            msg = msg + "\n%5d: %8s\t%8s\t\t%s\t%s %s" % (i, p_c1, p_c2, ord(c1), ord(c2), flag)
        else:
            # str1 is longer than str2
            flag = " <---"
            p_c1 = printable_for_char(c1)
            msg = msg + "\n%5d: %8s\t%8s\t\t%s\t%s %s" % (i, p_c1, '  ', ord(c1), '  ', flag)

    # TODO add code for str2 longer than str1

    return msg

# http://code.activestate.com/recipes/576874-levenshtein-distance/
def levenshtein(s1, s2):
    l1 = len(s1)
    l2 = len(s2)

    matrix = [range(l1 + 1)] * (l2 + 1)
    for zz in range(l2 + 1):
      matrix[zz] = range(zz,zz + l1 + 1)
    for zz in range(0,l2):
      for sz in range(0,l1):
        if s1[sz] == s2[zz]:
          matrix[zz+1][sz+1] = min(matrix[zz+1][sz] + 1, matrix[zz][sz+1] + 1, matrix[zz][sz])
        else:
          matrix[zz+1][sz+1] = min(matrix[zz+1][sz] + 1, matrix[zz][sz+1] + 1, matrix[zz][sz] + 1)
    return matrix[l2][l1]

def indent(s, spaces=4):
    return "\n".join("%s%s" % (' ' * spaces, line)
            for line in s.splitlines())



from dexy.filters.pydoc import PythonIntrospection
import io
import dexy.exceptions
import inspect
import os

try:
    import nose
    AVAILABLE = True
except ImportError:
    AVAILABLE = False

class PythonTest(PythonIntrospection):
    """
    Runs the tests in the specified Python modules.

    Python modules must be installed on the system. Returns a key-value store
    with test results and source code.
    
    Many packages are installed without tests, so this won't work.
    """
    aliases = ['pytest']
    _settings = {
            'run-tests' : (
                "Whether to run tests or just return test source code.",
                True
                ),
            'test-passed-when-not-run' : (
                "Value which should be used for test_passed when tests are not run.",
                True
                ),
            'load-tests-mode' : (
                "'dir' to find tests by location relative to package dir, 'name' if tests are in their own module.",
                'dir'),
            'path-to-tests-dir' : (
                "Path from package dir to tests dir.",
                "../tests"),
            'chdir' : (
                "Path from package dir to chdir to.",
                None),
            'nose-argv' : (
                "Need to fake out argv since sys.argv will be args for dexy, not nose.",
                ['nosetests', '--stop', '--verbose'])
            }

    def is_active(self):
        return AVAILABLE

    def load_tests_from_dir(self, module_name):
        self.log_debug("Loading module '%s' to find its tests." % module_name)
        mod = self.load_module(module_name)

        self.mod_file_dir = os.path.dirname(mod.__file__)
        relpath = self.setting('path-to-tests-dir')
        tests_dir = os.path.normpath(os.path.join(self.mod_file_dir, relpath))
        self.log_debug("Attempting to load tests from dir '%s'" % tests_dir)

        loader = nose.loader.TestLoader()
        return loader.loadTestsFromDir(tests_dir)

    def load_tests_from_name(self, module_name):
        loader = nose.loader.TestLoader()
        return loader.loadTestsFromName(module_name)

    def load_tests(self, module_name):
        mode = self.setting('load-tests-mode')
        if mode == 'dir':
            return self.load_tests_from_dir(module_name)
        elif mode == 'name':
            return self.load_tests_from_name(module_name)
        else:
            msg = "Invalid load-tests-mode setting '%s'" % mode
            raise dexy.exceptions.UserFeedback(msg)

    def append_source(self, test, test_passed):
        for key in dir(test.context):
            member = test.context.__dict__[key]

            if inspect.ismethod(member) or inspect.isfunction(member):
                qualified_test_name = "%s.%s" % (test.context.__name__, member.__name__)
                source = inspect.getsource(member.__code__)

                if member.func_doc:
                    doc = inspect.cleandoc(member.func_doc)
                    self.output_data.append("%s:doc" % qualified_test_name, doc)

                comments = inspect.getcomments(member.__code__)

                self.output_data.append("%s:source" % qualified_test_name, source)
                self.output_data.append("%s:name" % qualified_test_name, member.func_name)
                self.output_data.append("%s:comments" % qualified_test_name, comments)
                self.output_data.append("%s:passed" % qualified_test_name, str(test_passed))

    def run_test(self, test):
        # TODO This isn't working... maybe because we're running this in a test
        noselogs = io.StringIO()
        config = nose.config.Config(
                logStream = noselogs
                )

        if self.setting('run-tests'):
            self.log_debug("Running test suite %s" % test)
            test_passed = nose.core.run(
                    suite=test,
                    config=config,
                    argv=self.setting('nose-argv')
                    )
            self.log_debug("Passed: %s" % test_passed)
        else:
            test_passed = self.setting('test-passed-when-not-run')

        return test_passed

    def process(self):
        module_names = str(self.input_data).split()
        self.mod_file_dir = None
        orig_wd = os.path.abspath(".")
        chdir = self.setting('chdir')

        for module_name in module_names:
            tests = self.load_tests(module_name)

            if chdir:
                chdir = os.path.abspath(os.path.join(self.mod_file_dir, chdir))
                self.log_warn("Changing dir to %s for tests" % chdir)
                os.chdir(chdir)

            for test in tests:
                test_passed = self.run_test(test)
                self.append_source(test, test_passed)

            if chdir:
                self.log_warn("Changing dir back to %s" % orig_wd)
                os.chdir(orig_wd)

        self.output_data.save()



from dexy.filter import DexyFilter
import base64
import dexy.exceptions
import json
import urllib

try:
    import IPython.nbformat.current
    AVAILABLE = True
except ImportError:
    AVAILABLE = False

class IPythonBase(DexyFilter):
    """
    Base class for IPython filters which work by loading notebooks into memory.
    """
    aliases = []

    def is_active(self):
        return AVAILABLE

    def load_notebook(self):
        nb = None
        with open(self.input_data.storage.data_file(), "r") as f:
            nb_fmt = self.input_data.ext.replace(".","")
            nb = IPython.nbformat.current.read(f, nb_fmt)
        return nb

    def enumerate_cells(self, nb=None):
        if not nb:
            nb = self.load_notebook()
        worksheet = nb['worksheets'][0]
        for j, cell in enumerate(worksheet['cells']):
            yield(j, cell)

class IPythonExport(IPythonBase):
    """
    Generates a static file based on an IPython notebook.
    """
    aliases = ['ipynbx']

    _settings = {
            'added-in-version' : "0.9.9.6",
            'input-extensions' : ['.ipynb'],
            'output' : True,
            'output-extensions' : ['.md', '.html'], # TODO add other formats
            }

    def process_html(self):
        nb = self.load_notebook()

    def process_md(self):
        output = ""
        for j, cell in self.enumerate_cells():
            cell_type = cell['cell_type']
            if cell_type == 'heading':
                output += "## %s\n" % cell['source']
            elif cell_type == 'markdown':
                output += "\n%s\n" % cell['source']
            elif cell_type == 'code':
                for k, cell_output in enumerate(cell['outputs']):
                    cell_output_type = cell_output['output_type']
                    del cell_output['output_type']
                    if cell_output_type == 'stream':
                        output += cell_output['text']
                    elif cell_output_type in ('pyout', 'pyerr',):
                        pass
                    elif cell_output_type == 'display_data':
                        for fmt, contents in cell_output.items():
                            if fmt == "png":
                                cell_output_image_file = "cell-%s-output-%s.%s" % (j, k, fmt)
                                d = self.add_doc(cell_output_image_file, base64.decodestring(contents))
                                output += "\n![Description](%s)\n" % urllib.quote(cell_output_image_file)
                            elif fmt in ('metadata','text',):
                                pass
                            else:
                                raise dexy.exceptions.InternalDexyProblem(fmt)
                    else:
                        raise dexy.exceptions.InternalDexyProblem("unexpected cell output type %s" % cell_output_type)
            else:
                raise dexy.exceptions.InternalDexyProblem("Unexpected cell type %s" % cell_type)

        return output

    def process(self):
        if self.ext == '.html':
            output = self.process_html()
        elif self.ext == '.md':
            output = self.process_md()
        else:
            raise dexy.exceptions.InternalDexyProblem("Shouldn't get ext %s" % self.ext)

        self.output_data.set_data(output)

class IPythonNotebook(IPythonBase):
    """
    Get data out of an IPython notebook.
    """
    aliases = ['ipynb']

    _settings = {
            'added-in-version' : "0.9.9.6",
            'examples' : ['ipynb'],
            'input-extensions' : ['.ipynb', '.json', '.py'],
            'output-extensions' : ['.json'],
            }

    def process(self):
        output = {}
        nb = self.load_notebook()

        nb_fmt_string = "%s.%s" % (nb['nbformat'], nb['nbformat_minor']) # 3.0 currently
        output['nbformat'] = nb_fmt_string

        cells = []
        documents = []

        for j, cell in self.enumerate_cells(nb):
            # could also do: cell_key = "%s--%0.3d" % (self.input_data.rootname(), j)
            cell_key = "%s--%s" % (self.input_data.rootname(), j)
            cell_type = cell['cell_type']

            if cell_type == 'heading':
                # TODO implement
                # keys are [u'source', u'cell_type', u'level', u'metadata']
                pass

            elif cell_type == 'markdown':
                d = self.add_doc("%s.md" % cell_key, cell['source'], {'output':False})
                documents.append(d.key)
                d = self.add_doc("%s.md|pyg|h" % cell_key, cell['source'])
                d = self.add_doc("%s.md|pyg|l" % cell_key, cell['source'])

            elif cell_type == 'code':
                # keys are [u'cell_type', u'language', u'outputs', u'collapsed', u'prompt_number', u'input', u'metadata']

                # map languages to file extensions to create new doc(s) for each cell
                file_extensions = {
                    'python' : '.py'
                    }
                ext = file_extensions[cell['language']]

                d = self.add_doc("%s-input%s" % (cell_key, ext), cell['input'], {'output': False })
                documents.append(d.key)

                # Add pygments syntax highlighting in HTML and LaTeX formats.
                self.add_doc("%s-input%s|pyg|h" % (cell_key, ext), cell['input'], { 'output' : False })
                self.add_doc("%s-input%s|pyg|l" % (cell_key, ext), cell['input'], { 'output' : False })

                # process each output
                for k, cell_output in enumerate(cell['outputs']):
                    cell_output_type = cell_output['output_type']
                    del cell_output['output_type']

                    if cell_output_type == 'stream':
                        assert sorted(cell_output.keys()) == ["stream", "text"], "stream output keys"
                        d = self.add_doc(
                                "%s-output-%s.txt" % (cell_key, k),
                                cell_output['text'],
                                {'output' : False}
                                )
                        documents.append(d.key)

                    elif cell_output_type == 'pyout':
                        pass

                    elif cell_output_type == 'pyerr':
                        pass

                    elif cell_output_type == 'display_data':
                        for fmt, contents in cell_output.items():
                            if fmt == "png":
                                d = self.add_doc(
                                        "%s-output-%s.%s" % (cell_key, k, fmt),
                                        base64.decodestring(contents)
                                        )
                                documents.append(d.key)
                                cell.outputs[k]['png'] = d.key
                            elif fmt == 'text':
                                pass
                            elif fmt == 'metadata':
                                pass
                            elif fmt == 'latex':
                                pass

                            else:
                                raise Exception("unexpected format in display_data %s" % fmt)

                    else:
                        raise Exception("unexpected output type %s" % cell_output_type)
            else:
                raise Exception("unexpected cell type '%s'" % cell_type)

            cells.append((cell_type, cell,))

        output["nbformat"] = nb_fmt_string
        output["cells"] = cells
        output["documents"] = documents
        for k, v in nb['metadata'].items():
            output[k] = v

        self.output_data.set_data(json.dumps(output))



from dexy.doc import Doc
from dexy.wrapper import Wrapper
from nose.exc import SkipTest
from tests.plugins.test_pexpect_filters import SCALA
from tests.utils import TEST_DATA_DIR
from tests.utils import assert_in_output
from tests.utils import assert_output
from tests.utils import assert_output_cached
from tests.utils import wrap
import os
import shutil

C_HELLO_WORLD = """#include <stdio.h>

int main()
{
    printf("HELLO, world\\n");
}
"""

def test_mkdirs():
    with wrap() as wrapper:
        doc = Doc("hello.c|c",
                wrapper,
                contents = C_HELLO_WORLD,
                c = {'mkdir' : 'foo', 'mkdirs' : ['bar', 'baz']}
                )
        wrapper.run_docs(doc)
        dirs = os.listdir(doc.filters[-1].workspace())
        assert 'foo' in dirs
        assert 'bar' in dirs
        assert 'baz' in dirs

def test_taverna():
    raise SkipTest()
    with wrap() as wrapper:
        orig = os.path.join(TEST_DATA_DIR, 'simple_python_example_285475.t2flow')
        shutil.copyfile(orig, 'simple-python.t2flow')
        node = Doc("simple-python.t2flow|taverna",
                wrapper)
        wrapper.run_docs(node)

PYIN_CONTENTS = """import sys
i = 0
while True:
    i += 1
    line = sys.stdin.readline()
    if not line:
        break
    print("line %s has %s chars" % (i, len(line)))
"""

def test_scalac():
    assert_output('scala', SCALA, "Hello, world!\n", ext=".scala", basename="HelloWorld")

def test_python_input():
    with wrap() as wrapper:
        node = Doc("hello.py|pyin",
                wrapper,
                [
                    Doc("input.in",
                        wrapper,
                        [],
                        contents="here is some input\nmore")
                    ],
                contents=PYIN_CONTENTS
                )
        wrapper.run_docs(node)
        assert str(node.output_data()) == """\
line 1 has 19 chars
line 2 has 4 chars
"""

def test_pandoc_filter_odt():
    # TODO Why isn't this checking for inactive filters?
    with wrap() as wrapper:
        node = Doc("hello.md|pandoc",
                wrapper,
                [],
                contents = "hello",
                pandoc = { "ext" : ".odt"}
                )
        wrapper.run_docs(node)
        wrapper.report()
        assert os.path.exists("output/hello.odt")

def test_pandoc_filter_pdf():
    with wrap() as wrapper:
        node = Doc("hello.md|pandoc",
                wrapper,
                [],
                contents = "hello",
                pandoc = { "ext" : ".pdf"}
                )
        wrapper.run_docs(node)
        wrapper.report()
        assert os.path.exists("output/hello.pdf")

def test_pandoc_filter_txt():
    with wrap() as wrapper:
        node = Doc("hello.md|pandoc",
            wrapper, [],
            contents = "hello",
            pandoc = { "ext" : ".txt"},
            )
        wrapper.run_docs(node)
        wrapper.report()
        assert os.path.exists("output/hello.txt")
        assert str(node.output_data()) == 'hello\n\n'

R_SECTIONS = """\
### @export "assign-vars"
x <- 6
y <- 7

### @export "multiply"
x * y
"""

def test_rint_mock():
    with wrap() as wrapper:
        node = Doc("example.R|idio|rintmock",
                wrapper,
                [],
                contents=R_SECTIONS
                )

        wrapper.run_docs(node)
        assert node.output_data().is_cached()
        assert str(node.output_data()['assign-vars']) == "> x <- 6\n> y <- 7\n> \n"
        assert str(node.output_data()['multiply']) == "> x * y\n[1] 42\n> \n"

def test_ht_latex():
    assert_output_cached("htlatex", LATEX, ext=".tex")

def test_r_batch():
    assert_output('rout', 'print(1+1)', "[1] 2\n")

def test_r_int_batch():
    assert_output('rintbatch', '1+1', "> 1+1\n[1] 2\n> \n")

def test_ragel_ruby_filter():
    assert_in_output('rlrb', RAGEL, "_keys = _hello_and_welcome_key_offsets[cs]", ext=".rl")

def test_ps2pdf_filter():
    with wrap() as wrapper:
        node = Doc("hello.ps|ps2pdf",
                wrapper, [],
                contents = PS)
        wrapper.run_docs(node)
        assert node.output_data().is_cached()
        assert node.output_data().filesize() > 1000

def test_html2pdf_filter():
    assert_output_cached("html2pdf", "<p>hello</p>", min_filesize=1000)

def test_dot_filter():
    assert_output_cached("dot", "digraph { a -> b }", min_filesize=1000, ext=".dot")

def test_pdf2img_filter():
    with wrap() as wrapper:
        orig = os.path.join(TEST_DATA_DIR, 'color-graph.pdf')
        shutil.copyfile(orig, 'example.pdf')
        wrapper=Wrapper()
        node = Doc("example.pdf|pdf2img", wrapper)
        wrapper.run_docs(node)
        assert node.output_data().is_cached()
        assert node.output_data().filesize() > 1000

def test_pdf2jpg_filter():
    with wrap() as wrapper:
        orig = os.path.join(TEST_DATA_DIR, 'color-graph.pdf')
        shutil.copyfile(orig, 'example.pdf')
        wrapper=Wrapper()
        node = Doc("example.pdf|pdf2jpg", wrapper)

        wrapper.run_docs(node)
        assert node.output_data().is_cached()

def test_bw_filter():
    with wrap() as wrapper:
        orig = os.path.join(TEST_DATA_DIR, 'color-graph.pdf')
        shutil.copyfile(orig, 'example.pdf')
        wrapper=Wrapper()
        node = Doc("example.pdf|bw", wrapper)

        wrapper.run_docs(node)
        assert node.output_data().is_cached()

def test_pdfcrop_filter():
    with wrap() as wrapper:
        orig = os.path.join(TEST_DATA_DIR, 'color-graph.pdf')
        shutil.copyfile(orig, 'example.pdf')
        wrapper=Wrapper()
        node = Doc("example.pdf|pdfcrop|pdfinfo", wrapper)

        wrapper.run_docs(node)
        assert node.output_data().is_cached()

def test_asciidoc_filter():
    assert_in_output("asciidoc", "hello", """<div class="paragraph"><p>hello</p></div>""")

def test_pandoc_filter():
    assert_output("pandoc", "hello", "<p>hello</p>\n", ext=".md")

def test_espeak_filter():
    assert_output_cached("espeak", "hello", min_filesize = 1000)

PS = """%!PS
1.00000 0.99083 scale
/Courier findfont 12 scalefont setfont
0 0 translate
/row 769 def
85 {/col 18 def 6 {col row moveto (Hello World)show /col col 90 add def}
repeat /row row 9 sub def} repeat
showpage save restore"""

RD = """
 \\name{load}
     \\alias{load}
     \\title{Reload Saved Datasets}
     \description{
       Reload the datasets written to a file with the function
       \code{save}.
     }
"""

RAGEL = """%%{
  machine hello_and_welcome;
  main := ( 'h' @ { puts "hello world!" }
          | 'w' @ { puts "welcome" }
          )*;
}%%
  data = 'whwwwwhw'
  %% write data;
  %% write init;
  %% write exec;
"""

LATEX = """\
\documentclass{article}
\\title{Hello, World!}
\\begin{document}
\maketitle
Hello!
\end{document}
"""



from dexy.commands.utils import print_indented
from dexy.reporters.output import Output
from dexy.utils import file_exists
from dexy.utils import iter_paths
from dexy.utils import reverse_iter_paths
from functools import partial
from jinja2 import Environment
from jinja2 import FileSystemLoader
import dexy.data
import dexy.exceptions
import dexy.filters.templating_plugins
import inspect
import jinja2
import os
import posixpath
import urllib

class Website(Output):
    """
    Applies a template to create a website from your dexy output.

    Templates are applied to all files with .html extension which don't already
    contain "<head" or "<body" tags.
    """
    aliases = ['ws']
    _other_class_settings = {
        'doc' : {
            'apply-ws-to-content' : ("""
                If you want to put website-related content (like the link()
                function) in your content, set this to True so your content
                gets put through the jinja filter for the website reporter.
                """,
                False
            ),
            'apply-ws-to-content-var-start-string' : ("""
                Provide a custom jinja var-start-string to avoid clashes.
                """,
                None
            ),
            'apply-ws-to-content-var-end-string' : ("""
                Provide a custom jinja var-end-string to avoid clashes.
                """,
                None
                ),
            'ws-template' : (
                """
                Key of the template to apply for rendering in website.
                Setting of 'None' will use default template, 'False' will
                force no template to be used.
                """,
                None)
            },
        'data' : {
            'ws-template' : (
                """
                Key of the template to apply for rendering in website.
                Setting of 'None' will use default template, 'False' will
                force no template to be used.
                """,
                None)
            }
        }

    _settings = {
        "dir" : "output-site",
        "default-template" : ("Path to the default template to apply.", "_template.html"),
        "default" : False
    }

    def run(self, wrapper):
        self.wrapper=wrapper

        self.remove_reports_dir(self.wrapper, keep_empty_dir=True)
        self.create_reports_dir()

        self.setup()

        if self.wrapper.target:
            msg = "Not running website reporter because a target has been specified."
            self.log_warn(msg)
            return

        for doc in list(wrapper.nodes.values()):
            if self.should_process(doc):
                self.process_doc(doc)

        self.log_debug("finished")

    def setup(self):
        self.keys_to_outfiles = []
        self.locations = {}
        self.create_reports_dir()
        self.setup_navobj()

    def setup_navobj(self):
        self._navobj = self.create_navobj()

    def should_process(self, doc):
        if not doc.key_with_class() in self.wrapper.batch.docs:
            return False
        elif not doc.state in ('ran', 'consolidated'):
            return False
        elif not hasattr(doc, 'output_data'):
            return False
        elif not doc.output_data().output_name():
            return False
        elif not doc.output_data().is_canonical_output():
            msg = "skipping %s - not canonical"
            self.log_debug(msg % doc.key)
            return False
        else:
            return True

    def process_doc(self, doc):
        self.log_debug("processing %s" % doc.key)

        output_ext = doc.output_data().ext

        if output_ext == ".html":
            self.process_html(doc)

        elif isinstance(doc.output_data(), dexy.data.Sectioned):
            assert output_ext == ".json"
            self.apply_and_render_template(doc)

        else:
            self.write_canonical_data(doc)

    def process_html(self, doc):
        if doc.setting('ws-template') == False:
            self.log_debug("  ws-template is False for %s" % doc.key)
            self.write_canonical_data(doc)

        elif self.detect_html_header(doc) and not doc.setting('ws-template'):
            self.log_debug("  found html tag in output of %s" % doc.key)
            self.write_canonical_data(doc)

        else:
            self.apply_and_render_template(doc)

    def detect_html_header(self, doc):
        fragments = ('<html', '<body', '<head')
        return any(html_fragment
                      in str(doc.output_data())
                      for html_fragment in fragments)

    def create_navobj(self):
        navobj = Navigation()
        navobj.populate_lookup_table(self.wrapper.batch)
        navobj.walk()
        return navobj

    def template_file_and_path(self, doc):
        ws_template = doc.setting('ws-template')
        if ws_template and not isinstance(ws_template, bool):
            template_file = ws_template
        else:
            template_file = self.setting('default-template')

        template_path = None
        for subpath in reverse_iter_paths(doc.name):
            template_path = os.path.join(subpath, template_file)
            if file_exists(template_path):
                break

        if not template_path:
            msg = "no template path for %s" % doc.key
            raise dexy.exceptions.UserFeedback(msg)
        else:
            msg = "  using template %s for %s"
            msgargs = (template_path, doc.key)
            self.log_debug(msg % msgargs)

        return (template_file, template_path,)

    def jinja_environment(self, template_path, additional_args=None):
        """
        Returns jinja Environment object.
        """
        args = {
                'undefined' : jinja2.StrictUndefined
                }

        if additional_args:
            args.update(additional_args)

        env = Environment(**args)

        dirs = [".", os.path.dirname(__file__), os.path.dirname(template_path)]
        env.loader = FileSystemLoader(dirs)

        return env

    def apply_jinja_to_page_content(self, doc, env_data):
        args = {
                'undefined' : jinja2.StrictUndefined
                }

        keys = ['variable_start_string', 'variable_end_string',
                'block_start_string', 'block_end_string']


        for k in keys:
            setting_name = "apply-ws-to-content-%s" % k.replace("_", "-")
            if doc.safe_setting(setting_name):
                args[k] = doc.setting(setting_name)

        env = Environment(**args)

        self.log_debug("Applying jinja to doc content %s" % doc.key)
        try:
            content_template = env.from_string(str(doc.output_data()))
            return content_template.render(env_data)
        except Exception:
            self.log_debug("Template:\n%s" % str(doc.output_data()))
            self.log_debug("Env args:\n%s" % args)
            raise

    def template_environment(self, doc, template_path):
        raw_env_data = self.run_plugins()
        raw_env_data.update(self.website_specific_template_environment(doc.output_data(), {
            'template_source' : ("The directory containing the template file used.",
                template_path)
            }))

        env_data = dict((k, v[1]) for k, v in raw_env_data.items())

        if doc.safe_setting('apply-ws-to-content'):
            env_data['content'] = self.apply_jinja_to_page_content(doc, env_data)
        else:
            env_data['content'] = doc.output_data()

        return env_data

    def fix_ext(self, filename):
        basename, ext = os.path.splitext(filename)
        return "%s.html" % basename

    def apply_and_render_template(self, doc):
        template_info = self.template_file_and_path(doc)
        template_file, template_path = template_info
        env_data = self.template_environment(doc, template_path)

        self.log_debug("  creating jinja environment")
        env = self.jinja_environment(template_path)

        self.log_debug("  loading jinja template at %s" % template_path)
        template = env.get_template(template_path)
       
        output_file = self.fix_ext(doc.output_data().output_name())
        output_path = os.path.join(self.setting('dir'), output_file)

        try:
            os.makedirs(os.path.dirname(output_path))
        except os.error:
            pass

        self.log_debug("  writing to %s" % (output_path))
        template.stream(env_data).dump(output_path, encoding="utf-8")

    def help(self, data):
        nodoc = ('navobj', 'navigation',)
        print_indented("Website Template Environment Variables:", 4)
        print('')
        print_indented("Navigation and Content Related:", 6)
        env = self.website_specific_template_environment(data)
        for k in sorted(env):
            if k in nodoc:
                continue
            docs, value = env[k]
            print_indented("%s: %s" % (k, docs), 8)

        print('')
        navobj = env['navobj'][1]
        root = navobj.nodes['/']
        members = [(name, obj) for name, obj in inspect.getmembers(root)
                if not name.startswith('__')]

        print_indented("navobj Node attributes (using root node):", 6)
        print('')
        for member_name, member_obj in members:
            if not inspect.ismethod(member_obj):
                print_indented("%s: %r" % (member_name, member_obj), 8)
       
        print('')

        print_indented("navobj Node methods (using root node):", 6)
        print('')
        for member_name, member_obj in members:
            if inspect.ismethod(member_obj):
                print_indented("%s()" % (member_name), 8)
                print_indented("%s" % inspect.getdoc(member_obj), 10)
                print_indented("%s" % member_obj(), 10)
                print('')

        print('')
        print_indented("Variables From Plugins:", 6)
        env = self.run_plugins()
        for k in sorted(env):
            docs, value = env[k]
            print_indented("%s: %s" % (k, docs), 8)

        print('')
        print_indented("navobj Nodes:", 4)
        print_indented(navobj.debug(), 6)

    def website_specific_template_environment(self, data, initial_args=None):
        env_data = {}

        if initial_args:
            env_data.update(initial_args)

        current_dir = posixpath.dirname(data.output_name())
        parent_dir = os.path.split(current_dir)[0]

        env_data.update({
                'link' : ("Function to create link to other page.",
                    partial(self.link, data)),
                'section' : ("Function to create link to section on any page.",
                    partial(self.section, data)),
                'navigation' : ("DEPRECATED. 'navigation' object.",
                    {}),
                'nav' : ("The node for the current document's directory.",
                    self._navobj.nodes["/%s" % current_dir]),
                'root' : ("The root node of the navigation tree.",
                    self._navobj.nodes["/"]),
                'navobj' : ("DEPRECATED. Same as 'navtree'.",
                    self._navobj),
                'navtree' : ("The complete navigation tree for the website.",
                    self._navobj),
                'page_title' : ("Title of the current page.",
                    data.title()),
                'title' : ("Title of the current page.",
                    data.title()),
                'parent_dir' : ("The directory one level up from the document being processed.",
                    parent_dir),
                'current_dir' : ("The directory containing the document being processed.",
                    current_dir),
                's' : ("'%s' type data object representing the current doc." % data.alias,
                    data),
                'd' : ("'%s' type data object representing the current doc." % data.alias,
                    data),
                'source' : ("Output name of current doc.",
                    data.output_name()),
                'wrapper' : ("The current wrapper object.",
                    self.wrapper)
                })

        return env_data

    def section(self, data, section_name=None, url_base=None, link_text = None):
        """
        Returns an HTML link to section without needing to specify which
        document it is in (section name must be globally unique).
        """
        matching_nodes = self.wrapper.lookup_sections.get(section_name)

        if not matching_nodes:
            msg = "Trying to create a link in %s but no section found matching '%s'"
            msgargs = (data.key, section_name,)
            raise dexy.exceptions.UserFeedback(msg % msgargs)
        elif len(matching_nodes) > 1:
            # TODO make it an option to select a default where there is
            # more than one option
            msg = "Trying to create a link in %s to '%s' but multiple docs match."
            msgargs = (data.key, section_name,)
            raise dexy.exceptions.UserFeedback(msg % msgargs)

        assert len(matching_nodes) == 1
        link_to_data = matching_nodes[0]
        section = link_to_data[section_name]
        anchor = section['id']
        if not link_text:
            link_text = section_name

        return self.link_for(url_base, data.relative_path_to(link_to_data.output_name()), link_text, anchor)

    def link(self, data, doc_key, section_name=None, url_base=None, link_text = None, description=False):
        """
        Returns an HTML link to document, optionally with an anchor linking to section.
        """
        matching_nodes = self.wrapper.lookup_nodes.get(doc_key)

        if not matching_nodes:
            msg = "Trying to create a link in %s but no doc found matching '%s'"
            msgargs = (data.key, doc_key,)
            raise dexy.exceptions.UserFeedback(msg % msgargs)
        elif len(matching_nodes) > 1:
            # TODO make it an option to select a default where there is
            # more than one option
            msg = "Trying to create a link to '%s' but multiple docs match."
            msgargs = (doc_key,)
            raise dexy.exceptions.UserFeedback(msg % msgargs)

        assert len(matching_nodes) == 1
        link_to_data = matching_nodes[0]
        anchor = None

        if section_name:
            if section_name in list(link_to_data.keys()):
                section = link_to_data[section_name]
                anchor = section['id']
                if not link_text:
                    link_text = section_name
            else:
                msg = "Did not find section named '%s' in '%s'"
                msgargs = (section_name, doc_key)
                raise dexy.exceptions.UserFeedback(msg % msgargs)
        else:
            if not link_text:
                link_text = link_to_data.title()


        relative_link_to = data.relative_path_to(link_to_data.output_name())

        link_html = self.link_for(url_base, relative_link_to, link_text, anchor)

        if description and link_to_data.safe_setting('description'):
            return "%s\n<p>%s</p>" % (link_html, link_to_data.setting('description'))
        else:
            return link_html

    def link_for(self, url_base, link, link_text, anchor=None):
        if url_base:
            url = urllib.parse.urljoin(url_base, link)
        else:
            url = link
        if anchor:
            return """<a href="%s#%s">%s</a>""" % (url, anchor, link_text)
        else:
            return """<a href="%s">%s</a>""" % (url, link_text)

class Navigation(object):
    def __init__(self):
        self.lookup_table = {}
        self.nodes = {}
        self.root = None

    def populate_lookup_table(self, batch):
        for data in batch:
            if not data.output_name():
                continue

            parent_dir = "/" + os.path.dirname(data.output_name())
            if not parent_dir in self.lookup_table:
                self.lookup_table[parent_dir] = {'docs' : []}

            if data.is_canonical_output():
                self.lookup_table[parent_dir]['docs'].append(data)

            if data.is_index_page():
                self.lookup_table[parent_dir]['index-page'] = data

    def walk(self):
        """
        Build nodes dict from already-populated lookup table.
        """
        for path, info in self.lookup_table.items():
            parent = None
            ancestors = []
            level = 0
            for parent_path in iter_paths(path):
                if not parent_path in self.nodes:
                    node = Node(parent_path, parent, [])
                    node.level = level
                    self.nodes[parent_path] = node
                    if parent:
                        parent.children.append(node)

                    if not self.root and parent_path == '/':
                        self.root = node

                parent = self.nodes[parent_path]
                ancestors.append(parent)
                level += 1

            assert parent.location == path
            parent.ancestors = ancestors
            parent.docs = info['docs']
            if 'index-page' in info:
                parent.index_page = info['index-page']

    def debug(self):
        """
        Returns a dump of useful information.
        """
        info = []
        for path in sorted(self.nodes):
            node = self.nodes[path]

            info.append('')
            info.append("node: %s" % path)

            if node.index_page:
                info.append("  index-page:")
                info.append("    %s" % node.index_page.key)
                info.append('')
            else:
                info.append("  no index page.")
                info.append('')

            if node.docs:
                info.append("  docs:")
            for child in node.docs:
                info.append("    %s" % child.key)
            info.append('')

            if node.children:
                info.append("  children:")

            for child in node.children:
                info.append("    %s" % child.location)
            info.append('')
        return "\n".join(info)

class Node(object):
    """
    Wrapper class for a location in the site hierarchy.
    """
    def __init__(self, location, parent, children):
        if parent:
            assert isinstance(parent, Node)
        if children:
            assert isinstance(children[0], Node)
        self.location = location
        self.parent = parent
        self.children = children
        self.ancestors = []
        self.docs = []
        self.index_page = None

    def __lt__(self, other):
        return self.location < other.location

    def __repr__(self):
        return "Node(%s)" % self.location

    def has_children_with_index_pages(self):
        """
        Boolean value.
        """
        return any(node.index_page for node in self.children)

    def breadcrumbs(self, divider = " &gt; "):
        """
        Navigation breadcrumbs showing each parent directory.
        """
        return divider.join("""<a href="%s">%s</a>""" % (node.location, node.index_page.title()) for node in self.ancestors if node.index_page)



from dexy.filter import DexyFilter
import dexy.exceptions
import json
import logging
import markdown
import re

class MarkdownFilter(DexyFilter):
    """
    Runs a Markdown processor to convert markdown to HTML.

    Markdown extensions can be enabled in your config:
    http://packages.python.org/Markdown/extensions/index.html
    """
    aliases = ['markdown']
    _settings = {
            'examples' : ['markdown'],
            'input-extensions' : ['.*'],
            'output-extensions' : ['.html'],
            'extensions' : ("Which Markdown extensions to enable.", { 'toc' : {} }),
            }

    def capture_markdown_logger(self):
        markdown_logger = logging.getLogger('MARKDOWN')
        markdown_logger.addHandler(self.doc.wrapper.log.handlers[-1])

    def initialize_markdown(self, *additional_extensions):
        extension_configs = self.setting('extensions')
        extensions = list(extension_configs.keys())

        dbg = "Initializing Markdown with extensions: %s and extension configs: %s"
        self.log_debug(dbg % (json.dumps(extensions), json.dumps(extension_configs)))

        try:
            md = markdown.Markdown(
                    extensions=extensions + list(additional_extensions),
                    extension_configs=extension_configs)
        except ValueError as e:
            self.log_debug(e.message)
            if "markdown.Extension" in e.message:
                raise dexy.exceptions.UserFeedback("There's a problem with the markdown extensions you specified.")
            else:
                raise
        except KeyError as e:
            raise dexy.exceptions.UserFeedback("Couldn't find a markdown extension option matching '%s'" % e.message)

        return md

    def process_text(self, input_text):
        self.capture_markdown_logger()
        md = self.initialize_markdown()
        return md.convert(input_text)

class MarkdownSlidesFilter(MarkdownFilter):
    """
    Converts paragraphs to HTML and wrap each slide in a header and footer.
    """
    aliases = ['slides']

    _settings = {
            'extensions' : { 'nl2br' : {} },
            'added-in-version': "0.9.9.6",
            'examples' : ['slides'],
            'comment-char' : (
                "Lines starting with this comment char will not show up in slides.",
                ';'),
            'split' : (
                "String to use to split slides.",
                "\n\n\n" # e.g. 2 blank lines.
                ),
            'slide-header' : (
                "Content to prepend to start of each slide.",
                """<section class="slide">"""
                ),
            'slide-footer' : (
                "Content to append to end of each slide.",
                """</section>"""
                ),
            }

    def process_text(self, input_text):
        self.capture_markdown_logger()
        md = self.initialize_markdown()

        slides = ""
        comment_regexp = "^%s(.*)$" % self.setting('comment-char')

        for counter, slide in enumerate(input_text.split(self.setting('split'))):
            slide = re.sub(comment_regexp, "", slide, flags=re.MULTILINE)
            html = md.convert(slide)

            # Variables to make available for string interpolation in header and footer.
            interp = {
                    'number' : (counter+1)
                    }

            header = self.setting('slide-header') % interp
            footer = self.setting('slide-footer')% interp

            slide_text = "\n%s\n%s\n%s\n" % (header, html, footer)
            slides += slide_text

        return slides

class MarkdownSpeakerNotesFilter(MarkdownSlidesFilter):
    """
    Companion to slides filter which helps generate speaker notes.
    """
    aliases = ['speakernotes']

    _settings = {
            'added-in-version' : '1.0.7',
            'comment-markdown' : ("Comment chars ; are replaced with this markdown. Use None to hide comments.", "> *")
            }

    def process_text(self, input_text):
        self.capture_markdown_logger()
        md = self.initialize_markdown()

        if self.setting('comment-markdown') is None:
            comment_regexp = "^%s(.*)$" % self.setting('comment-char')
            comments_removed = re.sub(comment_regexp, "", input_text, flags=re.MULTILINE)
        else:
            comment_regexp = "^%s" % self.setting('comment-char')
            comments_removed = re.sub(comment_regexp, self.setting('comment-markdown'), input_text, flags=re.MULTILINE)

        return md.convert(comments_removed)



from dexy.filter import DexyFilter

class LyxJinjaFilter(DexyFilter):
    """
    Converts dexy:foo.txt|bar into << d['foo.txt|bar'] >>
    
    Makes it easier to compose documents with lyx and process them in dexy.
    This expects you to do doc.lyx|lyx|lyxjinja|jinja|latex
    """
    aliases = ['lyxjinja']
    _settings = {
            'input-extensions' : ['.tex'],
            'output-extensions' : ['.tex'],
            }

    def process_text(self, input_text):
        lines = []
        for line in input_text.splitlines():
            if line.startswith("dexy:"):
                _, clean_line = line.split("dexy:")
                if ":" in clean_line:
                    doc, section = clean_line.split(":")
                    lines.append("<< d['%s']['%s'] >>" % (doc, section,))
                else:
                    lines.append("<< d['%s'] >>" % clean_line)
            else:
                lines.append(line)
        return "\n".join(lines)



from dexy.filter import DexyFilter
from lib2to3.pytree import Leaf
import inspect
import lib2to3.pgen2.driver
import lib2to3.pytree
import os

driver_file = lib2to3.pgen2.driver.__file__
grammar_dir = os.path.abspath(os.path.join(os.path.dirname(driver_file), ".."))

class PyParse(DexyFilter):
    """
    Parses Python source code using lib2to3, without loading files, so doesn't
    cause side effects or care if imports work.
    """
    aliases = ['pyparse']

    _settings = {
            'data-type' : 'keyvalue',
            'whitespace-types' : ("Token types corresponding to whitespace.", (4, 5)),
            'string-types' : ("Token types corresponding to strings.", (3,)),
            'grammar-file' : ("Name or full path to file specifying Python language grammar.", "Grammar.txt")
            }

    def grammar_path(self):
        grammar_file = self.setting('grammar-file')
        if "/" in grammar_file:
            return grammar_file
        else:
            return os.path.join(grammar_dir, grammar_file)

    def setup_driver(self):
        self.grammar = lib2to3.pgen2.driver.load_grammar(self.grammar_path())
        self.driver = lib2to3.pgen2.driver.Driver(self.grammar, convert=lib2to3.pytree.convert)

    def type_repr(self, type_num):
        if not hasattr(self, "_type_reprs"):
            from lib2to3.pygram import python_symbols
            self._type_reprs = {}
            for name, val in python_symbols.__dict__.items():
                if type(val) == int:
                    self._type_reprs[val] = name
        return self._type_reprs.setdefault(type_num, type_num)

    def type_of(self, node):
        return self.type_repr(node.type)

    def eval_string(self, leaf):
        return eval(leaf.value).strip()

    def first_simple_statement_after_any_whitespace(self, parent):
        for child in parent.children:
            if child.type in self.setting('whitespace-types'):
                pass
            elif child.type in self.setting('string-types'):
                first_child = child
                return self.eval_string(first_child)
            elif self.type_of(child) == "simple_stmt":
                first_child = child.children[0]
                if isinstance(first_child, Leaf) and first_child.type in self.setting('string-types'):
                    return self.eval_string(first_child)

    def process_node(self, node, prefix):
        node_type = self.type_of(node)

        decorators = []
        if node_type == 'decorated':
            # are we a decorated function or a decorated class?
            for child in node.children:
                if self.type_of(child) == "decorator":
                    decorators.append(str(child).rstrip())
                elif self.type_of(child) == "decorators":
                    for decorator in child.children:
                        decorators.append(str(decorator).rstrip())
                else:
                    child_type = self.type_of(child)
                    if child_type == "funcdef":
                        self.process_function(node, prefix, decorators)
                    elif child_type == "classdef":
                        self.process_class(node, prefix, decorators)
                    else:
                        raise Exception("unknown decorated child type '%s'" % child_type)

        elif node_type == 'funcdef':
            return self.process_function(node, prefix, decorators)
        elif node_type == "classdef":
            return self.process_class(node, prefix, decorators)
        elif node_type == 'simple_stmt':
            return self.process_simple_stmt(node, prefix, decorators)
        else:
            return prefix

    def process_root(self, node, prefix):
        docstring = self.first_simple_statement_after_any_whitespace(node)
        if prefix is None:
            key = ":doc"
        elif "None" in prefix:
            key = ":doc"
        else:
            key = "%s:doc" % prefix
   
        not_already_in_keys = not key in list(self.output_data.keys())
        is_a_docstring = docstring is not None
        if not_already_in_keys and is_a_docstring:
            self.output_data.append(key, inspect.cleandoc(docstring))

        self.process_node(node, prefix)

    def process(self):
        self.func_name = None
        self.class_name = None
        self.setup_driver()

        text = str(self.input_data)

        if text == "None":
            self.output_data.set_data({})
        else:
            root = self.driver.parse_string(text)

            for node in root.children:
                self.process_root(node, None)

            self.output_data.save()

    def name_with_prefix(self, name, prefix):
        if prefix is None:
            return name
        else:
            return "%s.%s" % (prefix, name)

    def process_function(self, node, prefix, decorators):
        def process_func_part(state, part):
            if state == "def":
                assert isinstance(part, Leaf) and part.value == "def", part
                return "funcname"
    
            elif state == "funcname":
                assert isinstance(part, Leaf) and part.type == 1
                self.func_name = part.value

                name_with_prefix = self.name_with_prefix(self.func_name, prefix)
                raw_source = str(node).rstrip().splitlines()

                is_function_started = False
                leading_comments = []
                source_lines = []
                indent = None
                for line in raw_source:
                    is_comment = line.lstrip().startswith("#")
                    is_decorator = line.lstrip().startswith("@")
                    has_def = "def" in line

                    if has_def and indent is None:
                        indent = len(line) - len(line.lstrip())

                    if is_function_started:
                        source_lines.append(line)

                    elif not is_function_started and is_comment:
                        leading_comments.append(line)

                    elif not is_comment and (has_def or is_decorator):
                        is_function_started = True
                        source_lines.append(line)
                  
                # Strip trailing comments and empty lines.
                while source_lines[-1].lstrip().startswith("#") or not source_lines[-1].lstrip():
                    source_lines.pop()

                # Fix decorator indent.
                for i, line in enumerate(source_lines):
                    if not line.startswith("@"):
                        break
                    if len(line) - len(line.lstrip()) == 0 and indent > 0:
                        source_lines[i] = " " * indent + line

                self.output_data.append("%s:source" % name_with_prefix, "\n".join(source_lines))
                self.output_data.append("%s:decorators" % name_with_prefix, "\n".join(decorators))
                return "parameters"
    
            elif state == "parameters":
                assert self.type_of(part) == "parameters"
                return "colon"
    
            elif state == "colon":
                assert part.value == ":"
                return "body"
    
            elif state == "body":
                assert self.type_of(part) == "suite"
                docstring = self.first_simple_statement_after_any_whitespace(part)
                if docstring is not None:
                    self.output_data.append("%s:doc" % self.name_with_prefix(self.func_name, prefix), inspect.cleandoc(docstring))
    
            elif state is None:
                pass
    
            else:
                raise Exception("invalid state '%s'" % state)
    
        state = "def"
        for child in node.children:
            if self.type_of(child) in ("decorator", "decorators",):
                pass
            elif self.type_of(child) == "funcdef":
                for grandchild in child.children:
                    state = process_func_part(state, grandchild)
                    if state == None:
                        break
            else:
                state = process_func_part(state, child)
                if state is None:
                    break
    
        name_with_prefix = self.name_with_prefix(self.func_name, prefix)
        self.func_name = None
        return name_with_prefix
    
    def process_class(self, node, prefix, decorators):
        def process_class_part(state, part):
            if state == "class":
                assert isinstance(part, Leaf) and part.value == "class"
                return "classname"
    
            elif state == "classname":
                assert isinstance(part, Leaf) and part.type == 1
                self.class_name = part.value
                name_with_prefix = self.name_with_prefix(self.class_name, prefix)
                self.output_data.append("%s:source" % name_with_prefix, str(node).strip())
                return "parameters"
    
            elif state == "parameters":
                if isinstance(part, Leaf) and part.value == ":":
                    return "body"
                else:
                    # continue processing parameters
                    return "parameters"
    
            elif state == "body":
                assert self.type_of(part) == "suite"

                # Parse any Class docstring
                name_with_prefix = self.name_with_prefix(self.class_name, prefix)
                docstring = self.first_simple_statement_after_any_whitespace(part)
                if docstring is not None:
                    self.output_data.append("%s:doc" % name_with_prefix, inspect.cleandoc(docstring))
   
                # Parse class and instance methods
                for child in part.children:
                    if self.type_of(child) in ('funcdef', 'decorated'):
                        self.process_node(child, self.name_with_prefix(self.class_name, prefix))
    
            else:
                raise Exception("invalid state '%s'" % state)
    
        state = "class"
        for child in node.children:
            if self.type_of(child) in ("decorator", "decorators",):
                pass
            elif self.type_of(child) == "classdef":
                for grandchild in child.children:
                    state = process_class_part(state, grandchild)
                    if state == None:
                        break
            else:
                state = process_class_part(state, child)
                if state is None:
                    break
   
        name_with_prefix = self.name_with_prefix(self.class_name, prefix)
        self.class_name = None
        return name_with_prefix

    def process_simple_stmt(self, node, prefix, decorators):
        for child in node.children:
            child_type = self.type_of(child)
            if child_type == "expr_stmt":
                if isinstance(child.children[0], Leaf):
                    name_leaf = child.children[0]
                    name = name_leaf.value
                    self.output_data.append("%s:source" % name, str(child))
                else:
                    # print "complex expr_stmt"
                    pass
            else:
                #print "unknown child type '%s'" % child_type
                pass



from dexy.data import Generic
import xml.etree.ElementTree as ET

class EtreeData(Generic):
    """
    Expose etree method to query XML using ElementTree.
    """
    aliases = ['etree']

    def etree(self):
        """
        Returns a tree root object.
        """
        if not hasattr(self, '_etree_root'):
            self._etree_root = ET.fromstring(self.data())
        return self._etree_root



from dexy.parser import AbstractSyntaxTree
from tests.utils import wrap
import dexy.batch

def test_ast():
    with wrap() as wrapper:
        wrapper.nodes = {}
        wrapper.roots = []
        wrapper.batch = dexy.batch.Batch(wrapper)
        wrapper.filemap = wrapper.map_files()

        ast = AbstractSyntaxTree(wrapper)

        ast.add_node("abc.txt", foo='bar', contents = 'abc')
        ast.add_dependency("abc.txt", "def.txt")
        ast.add_node("def.txt", foo='baz', contents = 'def')

        assert ast.tree == ['doc:abc.txt']
        assert ast.args_for_node('doc:abc.txt')['foo'] == 'bar'
        assert ast.args_for_node('doc:def.txt')['foo'] == 'baz'
        assert ast.inputs_for_node('abc.txt') == ['doc:def.txt']
        assert not ast.inputs_for_node('def.txt')

        ast.walk()
        assert len(wrapper.roots) == 1
        assert len(wrapper.nodes) == 2



from dexy.doc import Doc
from tests.utils import assert_in_output
from tests.utils import wrap
from nose.exc import SkipTest

def test_shint_filter():
    with wrap() as wrapper:
        src = """
### @export "touch"
touch newfile.txt

### @export "ls"
ls
"""
        doc = Doc("example.sh|idio|shint|pyg",
                wrapper,
                [],
                contents = src)
        wrapper.run_docs(doc)

        assert list(doc.output_data().keys()) == ['1', 'touch', 'ls']

SCALA = """object HelloWorld {
    def main(args: Array[String]) {
      println("Hello, world!")
    }
  }
"""

def test_scala_repl():
    raise SkipTest()
    with wrap() as wrapper:
        doc = Doc("HelloWorld.scala|scalai",
                wrapper,
                [],
                contents = SCALA
                )
        wrapper.run_docs(doc)
        assert "defined module HelloWorld" in str(doc.output_data())

RUST = """fn main() {
    io::println("hello?");
}"""

def test_rust_interactive():
    raise SkipTest("Need to get rust interactive filter working.")
    with wrap() as wrapper:
        doc = Doc("example.rs|rusti",
                wrapper,
                [],
                contents = "1+1"
                )
        wrapper.run_docs(doc)
        assert "rusti> 1+1\n2" in str(doc.output_data())

def test_rust():
    with wrap() as wrapper:
        doc = Doc("example.rs|rustc",
                wrapper,
                [],
                contents = RUST
                )
        wrapper.run_docs(doc)
        assert str(doc.output_data()) == "hello?\n"

PYTHON_CONTENT = """
x = 6
y = 7
"""
def test_python_filter_record_vars():
    with wrap() as wrapper:
        doc = Doc("example.py|pycon",
                wrapper,
                [],
                pycon = { 'record-vars' :  True},
                contents = PYTHON_CONTENT
                )

        wrapper.run_docs(doc)
        assert "doc:example.py-vars.json" in wrapper.nodes

def test_matlab_filter():
    raise SkipTest()
    assert_in_output('matlabint', "fprintf (1, 'Hello, world\\n')\n", "< M A T L A B (R) >")

def test_clj_filter():
    assert_in_output('cljint', '1+1', "user=> 1+1")

def test_ksh_filter():
    assert_in_output('kshint', 'ls', "example.txt")

def test_php_filter():
    assert_in_output('phpint', '1+1', "php > 1+1")

def test_rhino_filter():
    assert_in_output('rhinoint', '1+1', "js> 1+1")

def test_irb_filter():
    assert_in_output('irb', "puts 'hello'", ">> puts 'hello'")

def test_pycon_filter_single_section():
    assert_in_output('pycon', "print 'hello'", ">>> print 'hello'")

def test_ipython_filter():
    assert_in_output('ipython', "print 'hello'", ">>> print 'hello'")

def test_r_filter():
    assert_in_output('r', '1+1', '> 1+1')

def test_pycon_filter():
    with wrap() as wrapper:
        src = """
### @export "vars"
x = 6
y = 7

### @export "multiply"
x*y

"""
        node = Doc("example.py|idio|pycon",
                wrapper,
                [],
                contents=src)

        wrapper.run_docs(node)

        assert list(node.output_data().keys()) == ['1', 'vars', 'multiply']
        assert str(node.output_data()['vars']) == """
>>> x = 6
>>> y = 7"""

        assert str(node.output_data()['multiply']) == """
>>> x*y
42"""




from dexy.parser import Parser
from dexy.utils import parse_json
from dexy.utils import parse_yaml
import dexy.exceptions
import re

class Yaml(Parser):
    """
    Parses YAML configs.
    """
    aliases = ["dexy.yaml", "docs.yaml"]

    def parse(self, directory, input_text):
        def parse_key_mapping(mapping):
            for original_node_key, v in mapping.items():
                # handle things which aren't nodes
                if original_node_key == 'defaults':
                    self.ast.default_args_for_directory.append((directory, v,))
                    continue

                # handle nodes
                original_file = original_node_key.split("|")[0]

                orig_exists = self.file_exists(directory, original_file) 
                star_in_key = "*" in original_node_key
                dot_in_key = "." in original_node_key
                pipe_in_key = "|" in original_node_key

                treat_key_as_bundle_name = not orig_exists and not star_in_key and not dot_in_key and not pipe_in_key

                if treat_key_as_bundle_name:
                    node_key = original_node_key
                else:
                    node_key = self.wrapper.join_dir(directory, original_node_key)

                # v is a sequence whose members may be children or kwargs
                if not v:
                    raise dexy.exceptions.UserFeedback("Empty doc config for %s" % node_key)

                if hasattr(v, 'keys'):
                    raise dexy.exceptions.UserFeedback("You passed a dict to %s, please pass a sequence" % node_key)

                siblings = []
                for element in v:
                    if hasattr(element, 'keys'):
                        # This is a dict of length 1
                        kk = list(element.keys())[0]
                        vv = element[kk]

                        if isinstance(vv, list):
                            # This is a sequence. It probably represents a
                            # child task but if starts with 'args' or if it
                            # matches a filter alias for the parent doc, then
                            # it is nested complex kwargs.
                            if kk == "args" or (kk in node_key.split("|")):
                                # nested complex kwargs
                                for vvv in vv:
                                    self.ast.add_node(node_key, **vvv)

                            else:
                                # child task. we note the dependency, add
                                # dependencies on prior siblings, and recurse
                                # to process the child.
                                self.ast.add_dependency(node_key, self.wrapper.join_dir(directory, kk))

                                if self.wrapper.siblings:
                                    for s in siblings:
                                        self.ast.add_dependency(self.wrapper.join_dir(directory, kk), s)
                                    siblings.append(self.wrapper.join_dir(directory, kk))

                                parse_key_mapping(element)

                        else:
                            # This is a key:value argument for this task
                            self.ast.add_node(node_key, **element)

                    else:
                        # This is a child task with no args, we only have to
                        # note the dependencies
                        self.ast.add_dependency(node_key, self.wrapper.join_dir(directory, element))
                        if self.wrapper.siblings:
                            for s in siblings:
                                self.ast.add_dependency(self.wrapper.join_dir(directory, element), s)
                            siblings.append(self.wrapper.join_dir(directory, element))

        def parse_keys(data, top=False):
            if hasattr(data, 'keys'):
                parse_key_mapping(data)
            elif isinstance(data, str):
                self.ast.add_node(self.wrapper.join_dir(directory, data))
            elif isinstance(data, list):
                if top:
                    self.ast.root_nodes_ordered = True
                for element in data:
                    parse_keys(element)
            else:
                raise Exception("invalid input %s" % data)

        config = parse_yaml(input_text)
        parse_keys(config, top=True)

class TextFile(Parser):
    """
    parses plain text configs
    """
    aliases = ["dexy.txt", "docs.txt"]

    def parse(self, directory, input_text):
        for line in input_text.splitlines():
            line = line.strip()

            # Throw away comments.
            if "#" in line:
                if line.startswith("#"):
                    line = ''
                else:
                    line = line.split("#", 0)

            if not re.match("^\s*$", line):
                if "{" in line:
                    # We have a task + some JSON arguments
                    key, raw_args = line.split("{", 1)
                    key = key.strip()
                    kwargs = parse_json("{" + raw_args)
                else:
                    key = line
                    kwargs = {}

                node_key = self.wrapper.join_dir(directory, key)
                self.ast.add_node(node_key, **kwargs)
                # all tasks already in the ast are children
                for child_key in list(self.ast.lookup_table.keys()):
                    child_node_key = self.wrapper.join_dir(directory, child_key)
                    self.ast.add_dependency(node_key, child_node_key)

class Original(Parser):
    """
    parses JSON config files like .dexy
    """
    aliases = ["dexy.json", "docs.json", ".dexy"]

    def parse(self, directory, input_text):
        data = parse_json(input_text)

        for task_key, v in data.items():
            self.ast.add_node(self.wrapper.join_dir(directory, task_key))

            for kk, vv in v.items():
                if kk == 'depends':
                    for child_key in vv:
                        self.ast.add_dependency(self.wrapper.join_dir(directory, task_key), self.wrapper.join_dir(directory, child_key))
                else:
                    task_kwargs = {kk : vv}
                    self.ast.add_node(self.wrapper.join_dir(directory, task_key), **task_kwargs)

        def children_for_allinputs(priority=None):
            children = []
            for k, v in self.ast.lookup_table.items():
                if 'allinputs' in v:
                    if priority:
                        k_priority = v.get('priority', 10)
                        if k_priority < priority:
                            children.append(k)
                else:
                    children.append(k)
            return children

        # Make another pass to implement 'allinputs'
        for task_key, kwargs in self.ast.lookup_table.items():
            if kwargs.get('allinputs', False):
                priority = kwargs.get('priority')
                for child_key in children_for_allinputs(priority):
                    # These keys are already adjusted for directory.
                    self.ast.add_dependency(task_key, child_key)



from dexy.filter import DexyFilter

try:
    import pynliner
    AVAILABLE = True
except ImportError:
    AVAILABLE = False

class PynlinerFilter(DexyFilter):
    """
    Filter which exposes pynliner for inlining CSS styles into HTML.
    """
    aliases = ['pyn', 'pynliner']

    def is_active(self):
        return AVAILABLE

    def process_text(self, input_text):
        return pynliner.fromString(input_text)



from dexy.batch import Batch
from dexy.commands.utils import init_wrapper
from dexy.commands.utils import print_indented
from dexy.commands.utils import print_rewrapped
from dexy.utils import defaults
from operator import attrgetter
import dexy.exceptions
import sys

### "info-keys"
info_attrs = [
        'name',
        'ext',
        'key'
        ]

info_methods = [
        'title',
        'basename',
        'filesize',
        'baserootname',
        'parent_dir',
        'long_name',
        'web_safe_document_key'
        ]

storage_methods = []
### @end

def links_command(
        **kwargs
        ):
    """
    Print list of links and sections found in dexy documents.
    """
    artifactsdir = kwargs.get('artifactsdir', defaults['artifacts_dir'])
    wrapper = init_wrapper(locals())
    batch = Batch.load_most_recent(wrapper)

    if not batch:
        print("you need to run dexy first")
        sys.exit(1)

    wrapper.setup_log()
    wrapper.batch = batch

    wrapper.add_lookups()

    if wrapper.lookup_nodes:
        print_indented("Nodes:")
    for label in sorted(wrapper.lookup_nodes):
        nodes = wrapper.lookup_nodes[label]
        if len(nodes) > 1:
            print('')
            print_indented("'%s'" % label, 2)
            print_indented("Multiple nodes match %s:" % label, 4)
            for node in nodes:
                print_indented(">> %r" % node, 6)
        elif len(nodes) == 0:
            print_indented("'%s'" % label, 2)
            print_indented("NO nodes match %s" % label, 4)
        else:
            node = nodes[0]
            print_indented("'%s'" % label, 2)
            print_indented("%r" % node, 4)
        print('')

    print('')

    if wrapper.lookup_sections:
        print_indented("Sections:")
    for label in sorted(wrapper.lookup_sections):
        node = wrapper.lookup_sections[label][0]
        print_indented("'%s'" % label, 2)
        print_indented("%r" % node, 4)
        print('')

### "info-com"
def info_command(
        __cli_options=False,
        expr="", # An expression partially matching document name.
        key="", # The exact document key.
        ws=False, # Whether to print website reporter keys and values.
        **kwargs
        ):
    """
    Prints metadata about a dexy document.

    Dexy must have already run successfully.

    You can specify an exact document key or an expression which matches part
    of a document name/key. The `dexy grep` command is available to help you
    search for documents and print document contents.
    """
    artifactsdir = kwargs.get('artifactsdir', defaults['artifacts_dir'])
    wrapper = init_wrapper(locals())
    wrapper.setup_log()
    batch = Batch.load_most_recent(wrapper)
    wrapper.batch = batch

    if expr:
        print("search expr:", expr)
        matches = sorted([data for data in batch if expr in data.key],
                key=attrgetter('key'))
    elif key:
        matches = sorted([data for data in batch if key == data.key],
                key=attrgetter('key'))
    else:
        raise dexy.exceptions.UserFeedback("Must specify either expr or key")

    for match in matches:
        print("")
        print("  Info for Document '%s'" % match.key)
        print("")
        print("  document output data type:", match.alias)
        print("")

        print_indented("settings:", 2)
        for k in sorted(match._instance_settings):
            if not k in ('aliases', 'help'):
                print_indented("%s: %s" % (k, match.setting(k)), 4)

        print("")
        print_indented("attributes:", 2)
        for fname in sorted(info_attrs):
            print_indented("%s: %s" % (fname, getattr(match, fname)), 4)
        print("")
    
        print_indented("methods:", 2)
        for fname in sorted(info_methods):
            print_indented("%s(): %s" % (fname, getattr(match, fname)()), 4)
        print("")

        if storage_methods:
            print_indented("storage methods:", 2)
            for fname in sorted(storage_methods):
                print_indented("%s(): %s" % (fname, getattr(match.storage, fname)), 4)
            print('')

        if ws:
            print_indented("website reporter methods:", 2)
            print('')
            reporter = dexy.reporter.Reporter.create_instance('ws')
            reporter.wrapper = wrapper
            reporter.setup_navobj()
            reporter.help(match)
            print('')
            print_indented("active template plugins are:", 2)
            print_indented(", ".join(reporter.setting('plugins')), 4)
            print('')


        else:
            print_indented("For website reporter tags, run this command with -ws option", 4)
            print('')


        print_rewrapped("""For more information about methods available on this
        data type run `dexy datas -alias %s`""" % match.alias)



from dexy.data import Generic
from dexy.storage import GenericStorage
try:
    import tables
    AVAILABLE = True
except ImportError:
    AVAILABLE = False

class H5(Generic):
    """
    Data type for reading HDF5 files using pytables.
    """
    aliases = ['h5']
    _settings = {
            'storage-type' : 'h5storage'
            }

    def is_active(self):
        return AVAILABLE

    def root(self):
        return self.data().root

    def walk_groups(self, path=None):
        if path:
            return self.data().walk_groups(path)
        else:
            return self.data().walk_groups()

    def walk_nodes(self, path=None, node_type=None):
        if path and node_type:
            return self.data().walk_nodes(path, node_type)
        elif path:
            return self.data().walk_nodes(path)
        else:
            return self.data().walk_nodes()

class H5Storage(GenericStorage):
    """
    Storage backend representing HDF5 files.
    """
    aliases = ['h5storage']

    def is_active(self):
        return AVAILABLE

    def read_data(self):
        return tables.open_file(self.data_file(read=True), "r")

if AVAILABLE:
    # Set custom exit hook so messages about closing files don't get printed to
    # stderr, per http://www.pytables.org/moin/UserDocuments/AtexitHooks
    def my_close_open_files(verbose):
        try:
            open_files = tables.file._open_files
            are_open_files = len(open_files) > 0
            if verbose and are_open_files:
                print("Closing remaining open files:", end=' ', file=sys.stderr)
            for fileh in open_files.keys():
                if verbose:
                    print("%s..." % (open_files[fileh].filename,), end=' ', file=sys.stderr)
                open_files[fileh].close()
                if verbose:
                    print("done", end=' ', file=sys.stderr)
            if verbose and are_open_files:
                print(file=sys.stderr)
        except Exception:
            pass

    import sys, atexit
    atexit.register(my_close_open_files, False)



from dexy.filter import DexyFilter

class FluidHtml(DexyFilter):
    """
    Wraps your text in HTML header/footer which includes Baseline CSS resets.
    Easy way to add styles (includes Pygments syntax highlighting).
    """
    aliases = ['easyhtml']
    _settings = {
            'input-extensions' : ['.html'],
            'output-extensions' : ['.html'],
            "css" : ("Custom CSS to include in header.", ""),
            "js" : ("Custom JS to include (please wrap in script tags).", ""),
            }

    def process_text(self, input_text):
        css = self.setting('css')
        if css:
            self.log_debug("custom css is %s" % css)

        js = self.setting('js')
        if js:
            self.log_debug("custom js is %s" % js)

        args = {
                'pygments_css' : PYGMENTS_CSS,
                'css_framework' : CSS_FRAMEWORK,
                'buttons' : CSS_BUTTONS,
                'custom_css' : css,
                'custom_js' : js,
                'content' : input_text
                }
        return """
<html>
    <head>
        <meta http-equiv="Content-type" content="text/html;charset=UTF-8" />
        <style type="text/css">
            %(css_framework)s
            %(buttons)s
            %(pygments_css)s

            /* custom css */
            %(custom_css)s
        </style>
        <!-- custom js -->
        %(custom_js)s
    </head>
    <body>
    <div id="content">
        <div class="g3">
%(content)s
        </div>
    </div>
    </body>
</html>
""" % args




PYGMENTS_CSS = """
.highlight .hll { background-color: #ffffcc }
.highlight .c { color: #888888 } /* Comment */
.highlight .err { color: #a61717; background-color: #e3d2d2 } /* Error */
.highlight .k { color: #008800; font-weight: bold } /* Keyword */
.highlight .cm { color: #888888 } /* Comment.Multiline */
.highlight .cp { color: #cc0000; font-weight: bold } /* Comment.Preproc */
.highlight .c1 { color: #888888 } /* Comment.Single */
.highlight .cs { color: #cc0000; font-weight: bold; background-color: #fff0f0 } /* Comment.Special */
.highlight .gd { color: #000000; background-color: #ffdddd } /* Generic.Deleted */
.highlight .ge { font-style: italic } /* Generic.Emph */
.highlight .gr { color: #aa0000 } /* Generic.Error */
.highlight .gh { color: #303030 } /* Generic.Heading */
.highlight .gi { color: #000000; background-color: #ddffdd } /* Generic.Inserted */
.highlight .go { color: #888888 } /* Generic.Output */
.highlight .gp { color: #555555 } /* Generic.Prompt */
.highlight .gs { font-weight: bold } /* Generic.Strong */
.highlight .gu { color: #606060 } /* Generic.Subheading */
.highlight .gt { color: #aa0000 } /* Generic.Traceback */
.highlight .kc { color: #008800; font-weight: bold } /* Keyword.Constant */
.highlight .kd { color: #008800; font-weight: bold } /* Keyword.Declaration */
.highlight .kn { color: #008800; font-weight: bold } /* Keyword.Namespace */
.highlight .kp { color: #008800 } /* Keyword.Pseudo */
.highlight .kr { color: #008800; font-weight: bold } /* Keyword.Reserved */
.highlight .kt { color: #888888; font-weight: bold } /* Keyword.Type */
.highlight .m { color: #0000DD; font-weight: bold } /* Literal.Number */
.highlight .s { color: #dd2200; background-color: #fff0f0 } /* Literal.String */
.highlight .na { color: #336699 } /* Name.Attribute */
.highlight .nb { color: #003388 } /* Name.Builtin */
.highlight .nc { color: #bb0066; font-weight: bold } /* Name.Class */
.highlight .no { color: #003366; font-weight: bold } /* Name.Constant */
.highlight .nd { color: #555555 } /* Name.Decorator */
.highlight .ne { color: #bb0066; font-weight: bold } /* Name.Exception */
.highlight .nf { color: #0066bb; font-weight: bold } /* Name.Function */
.highlight .nl { color: #336699; font-style: italic } /* Name.Label */
.highlight .nn { color: #bb0066; font-weight: bold } /* Name.Namespace */
.highlight .py { color: #336699; font-weight: bold } /* Name.Property */
.highlight .nt { color: #bb0066; font-weight: bold } /* Name.Tag */
.highlight .nv { color: #336699 } /* Name.Variable */
.highlight .ow { color: #008800 } /* Operator.Word */
.highlight .w { color: #bbbbbb } /* Text.Whitespace */
.highlight .mf { color: #0000DD; font-weight: bold } /* Literal.Number.Float */
.highlight .mh { color: #0000DD; font-weight: bold } /* Literal.Number.Hex */
.highlight .mi { color: #0000DD; font-weight: bold } /* Literal.Number.Integer */
.highlight .mo { color: #0000DD; font-weight: bold } /* Literal.Number.Oct */
.highlight .sb { color: #dd2200; background-color: #fff0f0 } /* Literal.String.Backtick */
.highlight .sc { color: #dd2200; background-color: #fff0f0 } /* Literal.String.Char */
.highlight .sd { color: #dd2200; background-color: #fff0f0 } /* Literal.String.Doc */
.highlight .s2 { color: #dd2200; background-color: #fff0f0 } /* Literal.String.Double */
.highlight .se { color: #0044dd; background-color: #fff0f0 } /* Literal.String.Escape */
.highlight .sh { color: #dd2200; background-color: #fff0f0 } /* Literal.String.Heredoc */
.highlight .si { color: #3333bb; background-color: #fff0f0 } /* Literal.String.Interpol */
.highlight .sx { color: #22bb22; background-color: #f0fff0 } /* Literal.String.Other */
.highlight .sr { color: #008800; background-color: #fff0ff } /* Literal.String.Regex */
.highlight .s1 { color: #dd2200; background-color: #fff0f0 } /* Literal.String.Single */
.highlight .ss { color: #aa6600; background-color: #fff0f0 } /* Literal.String.Symbol */
.highlight .bp { color: #003388 } /* Name.Builtin.Pseudo */
.highlight .vc { color: #336699 } /* Name.Variable.Class */
.highlight .vg { color: #dd7700 } /* Name.Variable.Global */
.highlight .vi { color: #3333bb } /* Name.Variable.Instance */
.highlight .il { color: #0000DD; font-weight: bold } /* Literal.Number.Integer.Long */
"""

CSS_FRAMEWORK = """
/*
    Fluid Baseline Grid v1.0.0
    Designed & Built by Josh Hopkins and 40 Horse, http://40horse.com
    Licensed under Unlicense, http://unlicense.org/

    Base stylesheet with CSS normalization, typographic baseline grid and progressive responsiveness
*/

/* HTML5 DECLARATIONS */
article, aside, details, figcaption, figure, footer, header, hgroup, menu, nav, section, dialog {display: block}
audio[controls],canvas,video {display: inline-block; *display: inline; zoom: 1}

/* BASE */
html {height: 100%; font-size: 100%; overflow-y: scroll; -webkit-text-size-adjust: 100%} /* Force scrollbar in non-IE and Remove iOS text size adjust without disabling user zoom */
body {margin: 0; min-height: 100%; -webkit-font-smoothing:antialiased; font-smoothing:antialiased; text-rendering:optimizeLegibility; background:url('../images/24px_grid_bg.gif') 0 1.1875em} /* Improve default text rendering, handling of kerning pairs and ligatures */

/* DEFAULT FONT SETTINGS */
/* 16px base font size with 150% (24px) friendly, unitless line height and margin for vertical rhythm */
/* Font-size percentage is based on 16px browser default size */
body, button, input, select, textarea {font: 100%/1.5 Georgia,Palatino,"Palatino Linotype",Times,"Times New Roman",serif; *font-size: 1em; color: #333} /* IE7 and older can't resize px based text */
p, blockquote, q, pre, address, hr, code, samp, dl, ol, ul, form, table, fieldset, menu, img {margin: 0 0 1.5em; padding: 0}

/* TYPOGRAPHY */
/* Composed to a scale of 12px, 14px, 16px, 18px, 21px, 24px, 36px, 48px, 60px and 72px */
h1, h2, h3, h4, h5, h6 {font-family:Futura, "Century Gothic", AppleGothic, sans-serif;color:#222;text-shadow:1px 1px 1px rgba(0,0,0,.10)}
h1 {margin: 0; font-size: 3.75em; line-height: 1.2em; margin-bottom: 0.4em} /* 60px / 72px */
h2 {margin: 0; font-size: 3em; line-height: 1em; margin-bottom: 0.5em} /* 48px / 48px */
h3 {margin: 0; font-size: 2.25em; line-height: 1.3333333333333333333333333333333em; margin-bottom: 0.6667em} /* 36px / 48px */ 
h4 {margin: 0; font-size: 1.5em; line-height: 1em; margin-bottom: 1em} /* 24px / 24px */
h5 {margin: 0; font-size: 1.3125em; line-height: 1.1428571428571428571428571428571em; margin-bottom: 1.1428571428571428571428571428571em} /* 21px / 24px */
h6 {margin: 0; font-size: 1.125em; line-height: 1.3333333333333333333333333333333em; margin-bottom: 1.3333333333333333333333333333333em} /* 18px / 24px */
p, ul, blockquote, pre, td, th, label {margin: 0; font-size: 1em; line-height: 1.5em; margin-bottom: 1.5em} /* 16px / 24px */
small, p.small {margin: 0; font-size: 0.875em; line-height: 1.7142857142857142857142857142857em; margin-bottom: 1.7142857142857142857142857142857em} /* 14px / 24px */

/* CODE */
pre {white-space: pre; white-space: pre-wrap; word-wrap: break-word} /* Allow line wrapping of 'pre' */
pre, code, kbd, samp {font-size: 1em; line-height: 1.5em; margin-bottom: 1.5em; font-family: Menlo, Consolas, 'DejaVu Sans Mono', Monaco, monospace}

/* TABLES */
table {border-collapse: collapse; border-spacing: 0; margin-bottom: 1.5em}
th {text-align: left}
tr, th, td {padding-right: 1.5em; border-bottom: 0 solid #333}

/* FORMS */
form {margin: 0}
fieldset {border: 0;padding: 0}
textarea {overflow: auto; vertical-align: top}
legend {*margin-left: -.75em}
button, input, select, textarea {vertical-align: baseline; *vertical-align: middle} /* IE7 and older */
button, input {line-height: normal; *overflow: visible}
button, input[type="button"], input[type="reset"], input[type="submit"] {cursor: pointer;-webkit-appearance: button}
input[type="checkbox"], input[type="radio"] {box-sizing: border-box}
input[type="search"] {-webkit-appearance: textfield; -moz-box-sizing: content-box; -webkit-box-sizing: content-box; box-sizing: content-box}
input[type="search"]::-webkit-search-decoration {-webkit-appearance: none}
button::-moz-focus-inner, input::-moz-focus-inner {border: 0; padding: 0}

/* QUOTES */
blockquote, q {quotes: none}
blockquote:before, blockquote:after, q:before, q:after {content: ''; content: none}
blockquote, q, cite {font-style: italic}
blockquote {padding-left: 1.5em; border-left: 3px solid #ccc}
blockquote > p {padding: 0}

/* LISTS */
ul, ol {list-style-position: inside; padding: 0}
li ul, li ol {margin: 0 1.5em}
dl dd {margin-left: 1.5em}
dt {font-family:Futura, "Century Gothic", AppleGothic, sans-serif}

/* HYPERLINKS */
a {text-decoration: none; color:#c47529}
a:hover {text-decoration: underline}
a:focus {outline: thin dotted}
a:hover, a:active {outline: none} /* Better CSS Outline Suppression */

/* MEDIA */
figure {margin: 0}
img, object, embed, video {max-width: 100%; _width: 100%} /* Fluid images */
img {border: 0; -ms-interpolation-mode: bicubic} /* Improve IE's resizing of images */
svg:not(:root) {overflow: hidden} /* Correct IE9 overflow */

/* ABBREVIATION */
abbr[title], dfn[title] {border-bottom: 1px dotted #333; cursor: help}

/* MARKED/INSERTED/DELETED AND SELECTED TEXT */
ins, mark {text-decoration: none}
mark {background: #c47529}
ins {background: #d49855}
del {text-decoration: line-through}
::-moz-selection {background: #c47529; color: #fff; text-shadow: none} /* selected text */
::selection {background: #c47529; color: #fff; text-shadow: none} /* selected text */

/* OTHERS */
strong, b, dt { font-weight: bold}
dfn {font-style: italic}
var, address {font-style: normal}
sub, sup {font-size: 75%; line-height: 0; position: relative; vertical-align: baseline} /* Position 'sub' and 'sup' without affecting line-height */
sup {top: -0.5em} /* Move superscripted text up */
sub {bottom: -0.25em} /* Move subscripted text down */
span.amp{font-family:Adobe Caslon Pro,Baskerville,"Goudy Old Style","Palatino","Palatino Linotype","Book Antiqua",Georgia,"Times New Roman",Times,serif;font-style:italic;font-size:110%;line-height:0;position:relative;vertical-align:baseline} /* Best available ampersand */

/* MICRO CLEARFIX HACK */
.cf:before, .cf:after {content:"";display:table} /* For modern browsers */
.cf:after {clear:both}
.cf {zoom:1} /* For IE 6/7 (trigger hasLayout) */

/* DEFAULT MOBILE STYLE */
body {width: 92%; margin: 0 auto} /* Center page without wrapper */
/* column grid */
.g1,.g2,.g3{display:block; position: relative; margin-left: 1%; margin-right: 1%}
/* 1 column grid */
.g1,.g2,.g3{width:98.0%}


/* media Queries

FOLDING FLUID GRID
< 767px         - 1-Column Fluid Grid
768px - 1023px  - 2-Column Fluid Grid
> 1024px            - 3-Column Fluid Grid
Change widths as necessary
------------------------------------------- */

/* MOBILE PORTRAIT */
@media only screen and (min-width: 320px) {
    body {
        
    }
}

/* MOBILE LANDSCAPE */
@media only screen and (min-width: 480px) {
    body {
        
    }
}

/* SMALL TABLET */
@media only screen and (min-width: 600px) {
    body {
        
    }
}

/* TABLET/NETBOOK */
@media only screen and (min-width: 768px) { 
    body {
        
    }
    
    /* COLUMN GRID */
    .g1,.g2,.g3 {display:inline; float: left}
    
    /* 2 COLUMN GRID */
    .g1 {width:48.0%}
    .g2 {width:48.0%}
    .g3 {width:98.0%}
}

/* LANDSCAPE TABLET/NETBOOK/LAPTOP */
@media only screen and (min-width: 1024px) { 
    body {

    }
    
    /* 3 COLUMN GRID */
    .g1 {width:31.333%}
    .g2 {width:64.667%;}
    .g3 {width:98.0%}
}

@media only screen and (min-width: 1280px) { 
/* DESKTOP */
        body {

    }
}

/* WIDESCREEN */
/* Increased body size for legibility */
@media only screen and (min-width: 1400px) { 
    body {font-size:116.75%; background:url('../images/28px_grid_bg.gif') 0 1.25em; max-width:1440px} /* 18.5px / 28px */
}


/* PRINT */
@media print {
  * {background: transparent !important; color: black !important; text-shadow: none !important; filter:none !important; -ms-filter: none !important} /* Black prints faster */
  a, a:visited {color: #444 !important; text-decoration: underline}
  a[href]:after {content: " (" attr(href) ")"}
  abbr[title]:after {content: " (" attr(title) ")"}
  .ir a:after, a[href^="javascript:"]:after, a[href^="#"]:after {content: ""}  /* Don't print links for images, javascript or internal links */
  pre, blockquote {border: 1px solid #999; page-break-inside: avoid; }
  thead {display: table-header-group; } /* Repeat header row at top of each printed page */
  tr, img {page-break-inside: avoid; }
  img {max-width: 100% !important; }
  @page {margin: 0.5cm}
  p, h2, h3 {orphans: 3; widows: 3}
  h2, h3{page-break-after: avoid}
}
"""

CSS_BUTTONS = """
/*
 * Copyright (c) 2013 Thibaut Courouble
 * http://www.cssflow.com
 *
 * Licensed under the MIT License:
 * http://www.opensource.org/licenses/mit-license.php
 */

.button {
  font: 10px/18px 'Lucida Grande', Arial, sans-serif;
  display: inline-block;
  vertical-align: top;
  position: relative;
  overflow: hidden;
  min-width: 96px;
  line-height: 30px;
  padding: 0 24px;
  font-size: 14px;
  color: white;
  text-align: center;
  text-decoration: none;
  text-shadow: 0 1px #154c86;
  background-color: #247edd;
  background-clip: padding-box;
  border: 1px solid;
  border-color: #1c65b2 #18589c #18589c;
  border-radius: 4px;
  -webkit-box-shadow: inset 0 1px rgba(255, 255, 255, 0.4), 0 1px 2px rgba(0, 0, 0, 0.2);
  box-shadow: inset 0 1px rgba(255, 255, 255, 0.4), 0 1px 2px rgba(0, 0, 0, 0.2);
  background-image: -webkit-linear-gradient(top, rgba(255, 255, 255, 0.3), rgba(255, 255, 255, 0) 50%, rgba(0, 0, 0, 0.12) 51%, rgba(0, 0, 0, 0.04));
  background-image: -moz-linear-gradient(top, rgba(255, 255, 255, 0.3), rgba(255, 255, 255, 0) 50%, rgba(0, 0, 0, 0.12) 51%, rgba(0, 0, 0, 0.04));
  background-image: -o-linear-gradient(top, rgba(255, 255, 255, 0.3), rgba(255, 255, 255, 0) 50%, rgba(0, 0, 0, 0.12) 51%, rgba(0, 0, 0, 0.04));
  background-image: linear-gradient(to bottom, rgba(255, 255, 255, 0.3), rgba(255, 255, 255, 0) 50%, rgba(0, 0, 0, 0.12) 51%, rgba(0, 0, 0, 0.04));
}
.button:before {
  content: '';
  position: absolute;
  top: -25%;
  bottom: -25%;
  left: -20%;
  right: -20%;
  border-radius: 50%;
  background: transparent;
  -webkit-box-shadow: inset 0 0 38px rgba(255, 255, 255, 0.5);
  box-shadow: inset 0 0 38px rgba(255, 255, 255, 0.5);
}
.button:hover {
  background-color: #1a74d3;
  text-decoration: none;
}
.button:active {
  color: rgba(255, 255, 255, 0.9);
  text-shadow: 0 -1px #154c86;
  background: #1f71c8;
  border-color: #113f70 #154c86 #1c65b2;
  -webkit-box-shadow: inset 0 1px 3px rgba(0, 0, 0, 0.2), 0 1px rgba(255, 255, 255, 0.4);
  box-shadow: inset 0 1px 3px rgba(0, 0, 0, 0.2), 0 1px rgba(255, 255, 255, 0.4);
  background-image: -webkit-linear-gradient(top, #1a5da5, #3a8be0);
  background-image: -moz-linear-gradient(top, #1a5da5, #3a8be0);
  background-image: -o-linear-gradient(top, #1a5da5, #3a8be0);
  background-image: linear-gradient(to bottom, #1a5da5, #3a8be0);
}
.button:active:before {
  top: -50%;
  bottom: -125%;
  left: -15%;
  right: -15%;
  -webkit-box-shadow: inset 0 0 96px rgba(0, 0, 0, 0.2);
  box-shadow: inset 0 0 96px rgba(0, 0, 0, 0.2);
}

.button-green {
  text-shadow: 0 1px #0d4d09;
  background-color: #1ca913;
  border-color: #147b0e #11640b #11640b;
}
.button-green:hover {
  background-color: #159b0d;
}
.button-green:active {
  text-shadow: 0 -1px #0d4d09;
  background: #189210;
  border-color: #093606 #0d4d09 #147b0e;
  background-image: -webkit-linear-gradient(top, #126d0c, #20c016);
  background-image: -moz-linear-gradient(top, #126d0c, #20c016);
  background-image: -o-linear-gradient(top, #126d0c, #20c016);
  background-image: linear-gradient(to bottom, #126d0c, #20c016);
}

.button-red {
  text-shadow: 0 1px #72100d;
  background-color: #cd1d18;
  border-color: #9f1713 #891310 #891310;
}
.button-red:hover {
  background-color: #c01511;
}
.button-red:active {
  text-shadow: 0 -1px #72100d;
  background: #b61a15;
  border-color: #5b0d0b #72100d #9f1713;
  background-image: -webkit-linear-gradient(top, #921511, #e4201b);
  background-image: -moz-linear-gradient(top, #921511, #e4201b);
  background-image: -o-linear-gradient(top, #921511, #e4201b);
  background-image: linear-gradient(to bottom, #921511, #e4201b);
}
"""



from dexy.filter import Filter
from dexy.utils import file_exists
import dexy.exceptions
import fnmatch
import os
import platform
import subprocess

class SubprocessFilter(Filter):
    """
    Parent class for all filters which use the subprocess module.
    """
    aliases = []
    _settings = {
            'args' : ("Arguments to be passed to the executable.", ''),
            'check-return-code' : ("Whether to look for nonzero return code.", True),
            'clargs' : ("Arguments to be passed to the executable (same as 'args').", ''),
            'command-string' : ("The full command string.", """%(prog)s %(args)s "%(script_file)s" %(scriptargs)s "%(output_file)s" """),
            'make-dummy-output' : ("Whether to make a dummy output file when one is not generated and add-new-files is True.", False),
            'env' : ("Dictionary of key-value pairs to be added to environment for runs.", {}),
            'executable' : ('The executable to be run', None),
            'initial-timeout' : ('', 10),
            'path-extensions' : ("strings to extend path with", []),
            'record-vars' : ("Whether to add code that will automatically record values of variables.", False),
            'scriptargs' : ("Arguments to be passed to the executable.", ''),
            'tags' : [],
            'timeout' : ('', 10),
            'use-wd' : ("Whether to use a custom working directory when running filter.", True),
            'version-command': ( "Command to call to return version of installed software.", None),
            'windows-version-command': ( "Command to call on windows to return version of installed software.", None),
            'write-stderr-to-stdout' : ("Should stderr be piped to stdout?", True),
            }

    def version_command(klass):
        if platform.system() == 'Windows':
            return klass.setting('windows-version-command') or klass.setting('version-command')
        else:
            return klass.setting('version-command')

    def version(klass):
        command = klass.version_command()
        if command:
            proc = subprocess.Popen(
                       command,
                       shell=True,
                       stdout=subprocess.PIPE,
                       stderr=subprocess.STDOUT
                   )

            stdout, stderr = proc.communicate()
            if proc.returncode > 0:
                return False
            else:
                return stdout.decode('utf-8').strip().split("\n")[0]

    def process(self):
        command = self.command_string()
        proc, stdout = self.run_command(command, self.setup_env())
        self.handle_subprocess_proc_return(command, proc.returncode, stdout)
        self.copy_canonical_file()

        if self.setting('add-new-files'):
            self.log_debug("adding new files found in %s for %s" % (self.workspace(), self.key))
            self.add_new_files()

    def command_string_args(self):
        return self.default_command_string_args()

    def default_command_string_args(self):
        args = {
                'args' : " ".join([self.setting('args'), self.setting('clargs')]),
                'prog' : self.setting('executable'),
                'script_file' : self.work_input_filename(),
                'output_file' : self.work_output_filename()
                }
        skip = ['args', 'clargs']
        args.update(self.setting_values(skip))
        return args

    def command_string(self):
        return self.setting('command-string') % self.command_string_args()

    def ignore_nonzero_exit(self):
        return self.doc.wrapper.ignore_nonzero_exit

    def clear_cache(self):
        self.output_data.clear_cache()

    def handle_subprocess_proc_return(self, command, exitcode, stderr, compiled=False):
        self.log_debug("exit code is: %s" % exitcode)
        if exitcode is None:
            raise dexy.exceptions.InternalDexyProblem("no return code, proc not finished!")
        elif exitcode == 127 and not compiled:
            raise dexy.exceptions.InactivePlugin(self.alias)
        elif exitcode != 0 and self.setting('check-return-code'):
            if self.ignore_nonzero_exit():
                self.log_warn("Nonzero exit status %s" % exitcode)
                self.log_warn("output from process: %s" % stderr)
            else:
                err_msg = "The command '%s' for %s exited with nonzero exit status %s." % (command, self.key, exitcode)
                if stderr:
                    err_msg += " Here is stderr:\n%s" % stderr
                self.output_data.clear_cache()
                raise dexy.exceptions.UserFeedback(err_msg)

    def setup_timeout(self):
        return self.setting('timeout')

    def setup_initial_timeout(self):
        return self.setting('initial-timeout')

    def setup_env(self):
        env = os.environ

        env.update(self.setting('env'))

        env['DEXY_ROOT'] = os.path.abspath(".")

        # Add parameters in wrapper's env dict
        if self.is_part_of_script_bundle():
            for key, value in self.script_storage().items():
                if key.startswith("DEXY_"):
                    self.log_debug("Adding %s to env value is %s" % (key, value))
                    env[key] = value

        # Add any path extensions to PATH
        if self.setting('path-extensions'):
            paths = [env['PATH']] + self.setting('path-extensions')
            env['PATH'] = ":".join(paths)

        return env

    def add_new_files(self):
        """
        Walk working directory and add a new dexy document for every newly
        created file found.
        """
        wd = self.workspace()
        self.log_debug("adding new files found in %s for %s" % (wd, self.key))

        add_new_files = self.setting('add-new-files')
        if isinstance(add_new_files, str):
            add_new_files = [add_new_files]

        exclude = self.setting('exclude-add-new-files')
        skip_dirs = self.setting('exclude-new-files-from-dir')

        if isinstance(exclude, str):
            raise dexy.exceptions.UserFeedback("exclude-add-new-files should be a list, not a string")

        new_files_added = 0
        for dirpath, subdirs, filenames in os.walk(wd):
            # Prune subdirs which match exclude.
            subdirs[:] = [d for d in subdirs if d not in skip_dirs]

            # Iterate over files in directory.
            for filename in filenames:
                filepath = os.path.normpath(os.path.join(dirpath, filename))
                relpath = os.path.relpath(filepath, wd)
                self.log_debug("Processing %s" % filepath)

                if relpath in self._files_workspace_populated_with:
                    # already have this file
                    continue

                if isinstance(add_new_files, list):
                    is_valid_file_extension = False
                    for pattern in add_new_files:
                        if "*" in pattern:
                            if fnmatch.fnmatch(relpath, pattern):
                                is_valid_file_extension = True
                                continue
                        else:
                            if filename.endswith(pattern):
                                is_valid_file_extension = True
                                continue

                    if not is_valid_file_extension:
                        msg = "Not adding filename %s, does not match patterns: %s"
                        args = (filepath, ", ".join(add_new_files))
                        self.log_debug(msg % args)
                        continue

                elif isinstance(add_new_files, bool):
                    if not add_new_files:
                        msg = "add_new_files method should not be called if setting is False"
                        raise dexy.exceptions.InternalDexyProblem(msg)
                    is_valid_file_extension = True

                else:
                    msg = "add-new-files setting should be list or boolean. Type is %s value is %s"
                    args = (add_new_files.__class__, add_new_files,)
                    raise dexy.exceptions.InternalDexyProblem(msg % args)

                # Check if should be excluded.
                skip_because_excluded = False
                for skip_pattern in exclude:
                    if skip_pattern in filepath:
                        msg = "skipping adding new file %s because it matches exclude %s"
                        args = (filepath, skip_pattern,)
                        self.log_debug(msg % args)
                        skip_because_excluded = True
                        continue

                if skip_because_excluded:
                    continue

                if not is_valid_file_extension:
                    raise Exception("Should not get here unless is_valid_file_extension")

                self.log_debug("Adding %s" % filepath)
                with open(filepath, 'rb') as f:
                    raw_contents = f.read()

                try:
                    contents = raw_contents.decode('utf-8')
                except UnicodeDecodeError:
                    contents = raw_contents

                self.add_doc(relpath, contents)
                new_files_added += 1

        if new_files_added > 10:
            self.log_warn("%s additional files added" % (new_files_added))

    def run_command(self, command, env, input_text=None):
        if self.setting('use-wd'):
            ws = self.workspace()
            if os.path.exists(ws):
                self.log_debug("already have workspace '%s'" % os.path.abspath(ws))
            else:
                self.populate_workspace()

        stdout = subprocess.PIPE

        if input_text:
            stdin = subprocess.PIPE
        else:
            stdin = None

        if self.setting('write-stderr-to-stdout'):
            stderr = subprocess.STDOUT
        else:
            stderr = subprocess.PIPE

        if self.setting('use-wd'):
            wd = self.parent_work_dir()
        else:
            wd = os.getcwd()

        self.log_debug("about to run '%s' in '%s'" % (command, os.path.abspath(wd)))
        proc = subprocess.Popen(command, shell=True,
                                    cwd=wd,
                                    stdin=stdin,
                                    stdout=stdout,
                                    stderr=stderr,
                                    env=env)

        if input_text:
            self.log_debug("about to send input_text '%s'" % input_text)
            stdout, stderr = proc.communicate(input_text.encode())
        else:
            stdout, stderr = proc.communicate()

        self.log_debug("stdout is '%s'" % stdout.decode('utf-8'))

        if stderr:
            self.log_debug("stderr is '%s'" % stderr.decode('utf-8'))

        return (proc, stdout.decode('utf-8'))

    def copy_canonical_file(self):
        canonical_file = os.path.join(self.workspace(), self.output_data.name)
        if not self.output_data.is_cached() and file_exists(canonical_file):
            self.output_data.copy_from_file(canonical_file)

        if self.setting('add-new-files') and self.setting('make-dummy-output'):
            if not self.output_data.is_cached():
                self.output_data.set_data("ok")
                self.output_data.update_settings({"canonical-output": False})

class SubprocessStdoutFilter(SubprocessFilter):
    """
    Runs a command and returns the resulting stdout.
    """
    _settings = {
            'write-stderr-to-stdout' : False,
            'require-output' : False,
            'command-string' : '%(prog)s %(args)s "%(script_file)s" %(scriptargs)s'
            }

    def process(self):
        command = self.command_string()
        proc, stdout = self.run_command(command, self.setup_env())
        self.handle_subprocess_proc_return(command, proc.returncode, stdout)
        self.output_data.set_data(stdout)

        if self.setting('add-new-files'):
            self.add_new_files()

class SubprocessCompileFilter(SubprocessFilter):
    """
    Compiles code and runs the compiled executable.
    """
    _settings = {
            'add-new-files' : False,
            'check-return-code' : False,
            'compiled-extension' : ("Extension which compiled files end with.", ".o"),
            'compiler-command-string' : (
                "Command string to call compiler.",
                "%(prog)s %(compiler_args)s %(script_file)s -o %(compiled_filename)s"
                ),
            'compiler-args' : ("Args to pass to compiler.", '')
            }

    def compile_command_string(self):
        args = self.default_command_string_args()
        args['compiler_args'] = self.setting('compiler-args')
        args['compiled_filename'] = self.compiled_filename()
        return self.setting('compiler-command-string') % args

    def compiled_filename(self):
        basename = os.path.basename(self.input_data.name)
        nameroot = os.path.splitext(basename)[0]
        return "%s%s" % (nameroot, self.setting('compiled-extension'))

    def run_command_string(self):
        args = self.default_command_string_args()
        args['compiled_filename'] = self.compiled_filename()
        return "./%(compiled_filename)s %(args)s" % args

    def process(self):
        env = self.setup_env()

        # Compile the code
        command = self.compile_command_string()
        proc, stdout = self.run_command(command, env)

        # test exitcode from the *compiler*
        self.handle_subprocess_proc_return(command, proc.returncode, stdout)

        # Run the compiled code
        command = self.run_command_string()
        proc, stdout = self.run_command(command, env)

        # This tests exitcode from the compiled script.
        self.handle_subprocess_proc_return(command, proc.returncode, stdout, compiled=True)

        self.output_data.set_data(stdout)

        if self.setting('add-new-files'):
            msg = "adding new files found in %s for %s" 
            msgargs = (self.workspace(), self.key)
            self.log_debug(msg % msgargs)
            self.add_new_files()

class SubprocessInputFilter(SubprocessFilter):
    """
    Runs code which expects stdin.
    """
    _settings = {
            'data-type' : 'sectioned',
            'check-return-code' : False,
            'write-stderr-to-stdout' : False
            }

    def process(self):
        command = self.command_string()

        inputs = list(self.doc.walk_input_docs())

        if len(inputs) == 1:
            doc = inputs[0]
            for section_name, section_input in doc.output_data().items():
                proc, stdout = self.run_command(command, self.setup_env(), str(section_input))
                self.output_data[section_name] = stdout
        else:
            for doc in inputs:
                proc, stdout = self.run_command(command, self.setup_env(), str(doc.output_data()))
                self.handle_subprocess_proc_return(command, proc.returncode, stdout)
                self.output_data[doc.key] = stdout

        self.output_data.save()

class SubprocessInputFileFilter(SubprocessFilter):
    """
    Runs code which expects input files.
    """
    _settings = {
            'data-type' : 'sectioned',
            'check-return-code' : False,
            'write-stderr-to-stdout' : False,
            'command-string' : """%(prog)s %(args)s %(input_text)s "%(script_file)s" """
            }

    def command_string_args(self, input_doc):
        args = self.default_command_string_args()
        args['input_text'] = input_doc.name
        return args
    
    def command_string_for_input(self, input_doc):
        return self.setting('command-string') % self.command_string_args(input_doc)

    def process(self):
        self.populate_workspace()

        for doc in self.doc.walk_input_docs():
            command = self.command_string_for_input(doc)
            proc, stdout = self.run_command(command, self.setup_env())
            self.handle_subprocess_proc_return(command, proc.returncode, stdout)
            self.output_data[doc.key] = stdout

        self.output_data.save()

class SubprocessCompileInputFilter(SubprocessCompileFilter):
    """
    Compiles code and runs executable with stdin.
    """
    _settings = {
            'data-type' : 'sectioned',
            'check-return-code' : False,
            'write-stderr-to-stdout' : False
            }

    def process(self):
        # Compile the code
        command = self.compile_command_string()
        proc, stdout = self.run_command(command, self.setup_env())
        self.handle_subprocess_proc_return(command, proc.returncode, stdout)

        command = self.run_command_string()

        inputs = list(self.doc.walk_input_docs())

        if len(inputs) == 1:
            doc = inputs[0]
            for section_name, section_input in doc.output_data().items():
                proc, stdout = self.run_command(command, self.setup_env(), section_input)
                self.handle_subprocess_proc_return(command, proc.returncode, stdout)
                self.output_data[section_name] = stdout
        else:
            for doc in inputs:
                proc, stdout = self.run_command(command, self.setup_env(), str(doc.output_data()))
                self.handle_subprocess_proc_return(command, proc.returncode, stdout)
                self.output_data[doc.key] = stdout

        self.output_data.save()

class SubprocessFormatFlagFilter(SubprocessFilter):
    """
    Special handling of format flags based on file extensions.

    For example, ragel -R for ruby.
    """
    _settings = {
            'ext-to-format' : ("A dict of mappings from file extensions to format flags that need to be passed on the command line, e.g. for ragel with ruby host language .rb => -R", {})
            }

    def command_string_args(self):
        args = self.default_command_string_args()

        flags = self.setting('ext-to-format')

        if any(f in args['args'] for f in flags):
            # Already have specified the format manually.
            fmt = ''
        else:
            fmt = flags[self.ext]

        args['format'] = fmt
        return args

class SubprocessExtToFormatFilter(SubprocessFilter):
    """
    Subprocess filters which have ext-to-format param.
    """
    _settings = {
            'format-specifier' : ("The string used to specify the format switch, include trailing space if needed.", None),
            'ext-to-format' : ("A dict of mappings from file extensions to format parameters that need to be passed on the command line, e.g. for ghostscript .png => png16m", {})
            }

    def command_string_args(self):
        args = self.default_command_string_args()

        fmt_specifier = self.setting('format-specifier')
        if fmt_specifier and (fmt_specifier in args['args']):
            # Already have specified the format manually.
            fmt = ''
        else:
            fmt_setting = self.setting('ext-to-format')[self.ext]
            if fmt_setting:
                fmt = "%s%s" % (fmt_specifier, fmt_setting)
            else:
                fmt = ''

        args['format'] = fmt
        return args

class SubprocessStdoutTextFilter(SubprocessStdoutFilter):
    """
    Runs command with input passed on command line.
    """
    _settings = {
            'command-string' : "%(prog)s %(args)s \"%(text)s\"",
            'input-extensions' : ['.txt'],
            'output-extensions' : ['.txt']
            }

    def command_string_args(self):
        args = self.default_command_string_args()
        args['text'] = str(self.input_data)
        return args



from dexy.parser import AbstractSyntaxTree
from dexy.parsers.doc import Original
from dexy.parsers.doc import TextFile
from dexy.parsers.doc import Yaml
from tests.utils import wrap
from dexy.wrapper import Wrapper
import dexy.exceptions
import os

def test_text_parser():
    with wrap() as wrapper:
        with open("f1.py", "w") as f:
            f.write("print 'hello'")

        with open("f2.py", "w") as f:
            f.write("print 'hello'")

        with open("index.md", "w") as f:
            f.write("")

        wrapper = Wrapper()
        wrapper.to_valid()

        wrapper.nodes = {}
        wrapper.roots = []
        wrapper.batch = dexy.batch.Batch(wrapper)
        wrapper.filemap = wrapper.map_files()

        ast = AbstractSyntaxTree(wrapper)
        parser = TextFile(wrapper, ast)
        parser.parse(".", """
        *.py
        *.py|pyg
        *.md|jinja
        """)
        ast.walk()
        assert len(wrapper.nodes) == 8

YAML_WITH_INACTIVE = """
foo:
    - inactive: True
"""

def test_parse_inactive():
    with wrap() as wrapper:
        wrapper.nodes = {}
        wrapper.roots = []
        wrapper.batch = dexy.batch.Batch(wrapper)
        wrapper.filemap = wrapper.map_files()

        ast = AbstractSyntaxTree(wrapper)
        parser = Yaml(wrapper, ast)
        parser.parse('.', YAML_WITH_INACTIVE)
        ast.walk()
        assert len(wrapper.nodes) == 0

YAML_WITH_DEFAULT_OFF = """
foo:
    - default: False
"""

def test_parse_default():
    with wrap() as wrapper:
        wrapper.nodes = {}
        wrapper.roots = []
        wrapper.batch = dexy.batch.Batch(wrapper)
        wrapper.filemap = wrapper.map_files()

        ast = AbstractSyntaxTree(wrapper)
        parser = Yaml(wrapper, ast)
        parser.parse('.', YAML_WITH_DEFAULT_OFF)
        ast.walk()
        assert len(wrapper.nodes) == 0

    with wrap() as wrapper:
        wrapper.nodes = {}
        wrapper.roots = []
        wrapper.batch = dexy.batch.Batch(wrapper)
        wrapper.filemap = wrapper.map_files()

        ast = AbstractSyntaxTree(wrapper)
        wrapper.full = True
        parser = Yaml(wrapper, ast)
        parser.parse('.', YAML_WITH_DEFAULT_OFF)
        ast.walk()
        assert len(wrapper.nodes) == 1

def test_yaml_with_defaults():
    with wrap() as wrapper:
        os.makedirs("s1/s2")

        with open("s1/s2/hello.txt", "w") as f:
            f.write("hello")

        wrapper.nodes = {}
        wrapper.roots = []
        wrapper.batch = dexy.batch.Batch(wrapper)
        wrapper.filemap = wrapper.map_files()

        ast = AbstractSyntaxTree(wrapper)
        parser = Yaml(wrapper, ast)
        parser.parse('.', YAML_WITH_DEFAULTS)
        ast.walk()

        assert wrapper.roots[0].args['foo'] == 'bar'
    
def test_invalid_yaml():
    with wrap() as wrapper:
        ast = AbstractSyntaxTree(wrapper)
        parser = Yaml(wrapper, ast)
        try:
            parser.parse('.', INVALID_YAML)
            assert False, "should raise UserFeedback"
        except dexy.exceptions.UserFeedback as e:
            assert 'YAML' in e.message

def test_yaml_parser():
    with wrap() as wrapper:
        wrapper.nodes = {}
        wrapper.roots = []
        wrapper.batch = dexy.batch.Batch(wrapper)
        wrapper.filemap = wrapper.map_files()

        ast = AbstractSyntaxTree(wrapper)
        parser = Yaml(wrapper, ast)
        parser.parse('.', YAML)
        for doc in wrapper.roots:
            assert doc.__class__.__name__ == 'BundleNode'
            assert doc.key in ['code', 'wordpress']

        wrapper.run_docs()

def test_text_parser_blank_lines():
    with wrap() as wrapper:
        wrapper.nodes = {}
        wrapper.roots = []
        wrapper.batch = dexy.batch.Batch(wrapper)
        wrapper.filemap = wrapper.map_files()

        ast = AbstractSyntaxTree(wrapper)
        parser = TextFile(wrapper, ast)
        parser.parse('.', "\n\n")
        ast.walk()
        docs = wrapper.roots
        assert len(docs) == 0

def test_text_parser_comments():
    with wrap() as wrapper:
        wrapper.nodes = {}
        wrapper.roots = []
        wrapper.batch = dexy.batch.Batch(wrapper)
        wrapper.filemap = wrapper.map_files()

        ast = AbstractSyntaxTree(wrapper)
        parser = TextFile(wrapper, ast)
        parser.parse('.', """
        valid.doc { "contents" : "foo" }
        # commented-out.doc
        """)
        ast.walk()

        assert len(wrapper.roots) == 1
        assert wrapper.roots[0].key == "valid.doc"

def test_text_parser_valid_json():
    with wrap() as wrapper:
        wrapper.nodes = {}
        wrapper.roots = []
        wrapper.batch = dexy.batch.Batch(wrapper)
        wrapper.filemap = wrapper.map_files()

        ast = AbstractSyntaxTree(wrapper)
        parser = TextFile(wrapper, ast)
        parser.parse('.', """
        doc.txt { "contents" : "123" }
        """)
        ast.walk()

        docs = wrapper.roots
        assert docs[0].key == "doc.txt"
        assert docs[0].args['contents'] == "123"

def test_text_parser_invalid_json():
    with wrap() as wrapper:
        ast = AbstractSyntaxTree(wrapper)
        parser = TextFile(wrapper, ast)
        try:
            parser.parse('.', """
            doc.txt { "contents" : 123
            """)
            assert False, 'should raise UserFeedback'
        except dexy.exceptions.UserFeedback as e:
            assert 'unable to parse' in e.message

def test_text_parser_virtual_file():
    with wrap() as wrapper:
        wrapper.nodes = {}
        wrapper.roots = []
        wrapper.batch = dexy.batch.Batch(wrapper)
        wrapper.filemap = wrapper.map_files()

        ast = AbstractSyntaxTree(wrapper)
        parser = TextFile(wrapper, ast)
        parser.parse('.', """
        virtual.txt { "contents" : "hello" }
        """)
        ast.walk()

        wrapper.transition('walked')
        wrapper.to_checked()

        wrapper.run()
        docs = wrapper.roots

        assert docs[0].key == "virtual.txt"
        assert str(docs[0].output_data()) == "hello"

def test_original_parser():
    with wrap() as wrapper:
        wrapper.nodes = {}
        wrapper.roots = []
        wrapper.batch = dexy.batch.Batch(wrapper)
        wrapper.filemap = wrapper.map_files()

        conf = """{
        "*.txt" : {}
        }"""
        ast = AbstractSyntaxTree(wrapper)
        parser = Original(wrapper, ast)
        parser.parse('.', conf)
        ast.walk()

        assert wrapper.roots[0].key_with_class() == "pattern:*.txt"

def test_original_parser_allinputs():
    with wrap() as wrapper:
        wrapper.nodes = {}
        wrapper.roots = []
        wrapper.batch = dexy.batch.Batch(wrapper)
        wrapper.filemap = wrapper.map_files()

        conf = """{
        "*.txt" : {},
        "hello.txt" : { "contents" : "Hello!" },
        "*.md|jinja" : { "allinputs" : true }
        }"""

        ast = AbstractSyntaxTree(wrapper)
        parser = Original(wrapper, ast)
        parser.parse('.', conf)
        ast.walk()

        assert len(wrapper.roots) == 1
        assert wrapper.roots[0].key_with_class() == "pattern:*.md|jinja"

INVALID_YAML = """
code:
    - abc
    def
"""

YAML = """
code:
    - .R|pyg:
         "pyg" : { "foo" : "bar" }
    - .R|idio

wordpress:
    - code
    - test.txt|jinja:
        - contents: 'test'
    - .md|jinja|markdown|wp
"""

YAML_WITH_DEFAULTS = """
defaults:
    pyg: { lexer : moin }
    foo: bar

code:
    - .R|pyg

s1/s2/hello.txt|jinja:
    code
"""



from tests.utils import assert_output

def test_bleach_filter():
    assert_output("bleach", "an <script>evil()</script> example", 'an &lt;script&gt;evil()&lt;/script&gt; example')
    assert_output("bleach", "an <script>evil()</script> example", 'an &lt;script&gt;evil()&lt;/script&gt; example', ext=".html")



from tests.utils import assert_output_matches
from mock import patch, call
from io import StringIO


## See http://www.voidspace.org.uk/python/mock/patch.html#where-to-patch for
## why this mock applies to dexy.filters.websequence not urllib2
#@patch("dexy.filters.websequence.urlopen",
#       side_effect=[StringIO("?png=whereToGo"), StringIO("mock_result")])
#def test_websequence_svg(mocked_urlopen):
#    assert_output_matches("wsd", "Sender->Receiver", "mock_result", ext=".wsd")
#    print(mocked_urlopen.call_args_list)
#    assert mocked_urlopen.call_args_list[0] == call(
#        "http://www.websequencediagrams.com/index.php",
#        'message=Sender-%3EReceiver&style=default&apiVersion=1')
#    assert mocked_urlopen.call_args_list[1] == call(
#        "http://www.websequencediagrams.com/?png=whereToGo")



from dexy.exceptions import InternalDexyProblem
from dexy.exceptions import UserFeedback
from dexy.filter import DexyFilter
import imp
import inspect
import json
import os
import pkgutil
import sqlite3
import sys


# from https://nose.readthedocs.org/en/latest/plugins/skip.html
try:
    # 2.7
    from unittest.case import SkipTest
except ImportError:
    # 2.6 and below
    class SkipTest(Exception):
        """Raise this exception to mark a test as skipped.
        """
    pass

class add_workspace_to_sys_path(object):
    def __init__(self, *dirs):
        self.additional_dirs = dirs

    def __enter__(self):
        for dirname in self.additional_dirs:
            sys.path.append(dirname)

    def __exit__(self, type, value, traceback):
        for dirname in self.additional_dirs:
            sys.path.remove(dirname)

class PythonIntrospection(DexyFilter):
    """
    Base class for classes which use python introspection.
    """
    import_err_msg = "Could not import '%s' received err: %s"
    aliases = []
    _settings = {
            'error-on-import-fail' : (
                "Should an exception be raised if importing a specified module or file fails?",
                False
                ),
            'input-extensions' : ['.txt', '.py'],
            'skip-setup-py' : ("Whether to skip files named setup.py", True),
            'data-type' : 'keyvalue',
            'output-extensions' : ['.sqlite3', '.json']
            }

    def handle_fail(self, name, e):
        msg = self.import_err_msg % (name, e)
        if self.setting('error-on-import-fail'):
            raise UserFeedback(msg)
        else:
            self.log_debug(e)

    def load_module(self, name):
        try:
            __import__(name)
            return sys.modules[name]
        except (ImportError, TypeError) as e:
            self.handle_fail(name, e)

    def load_source_file(self):
        self.populate_workspace()
        with add_workspace_to_sys_path(self.workspace(), self.parent_work_dir()):
            name = self.input_data.name
            target = os.path.join(self.workspace(), name)

            # FIXME maybe make this more specific, search for certain text in setup.py like "setuptools"
            if name == "setup.py" and self.setting('skip-setup-py'):
                self.log_info("Skipping file %s because skip-setup-py is true." % target)
                return None

            try:
                return imp.load_source("dummy", target)
            except (ImportError, SkipTest) as e:
                self.handle_fail(name, e)

class Pydoc(PythonIntrospection):
    """
    Returns introspected python data in key-value storage format.

    Where input is a .txt file, this is assumed to be the name of an installed
    python module.

    Where input is a .py file, the file itself is loaded and parsed.
    """
    aliases = ["pydoc"]
    _settings = {
            'additional-dirs' : ("Additional source directories to load, relative to package root. Useful for tests/", [])
            }

    def append_item_content(self, key, item):
        self.log_debug("appending content for %s" % key)

        try:
            source = inspect.getsource(item)
            self.output_data.append("%s:source" % key, source)
        except (TypeError, IOError, sqlite3.ProgrammingError):
            pass

        try:
            doc = inspect.getdoc(item)
            self.output_data.append("%s:doc" % key, doc)
        except (TypeError, IOError, sqlite3.ProgrammingError):
            pass

        try:
            comment = inspect.getcomments(item)
            self.output_data.append("%s:comments" % key, comment)
        except (TypeError, IOError, sqlite3.ProgrammingError):
            pass

        try:
            value = json.dumps(item)
            self.output_data.append("%s:value" % key, value)
        except TypeError:
            pass

    def is_defined_in_module(self, mod, mod_name, item):
        if mod_name and hasattr(item, '__module__'):
            return item.__module__.startswith(mod_name)
        else:
            return True

    def process_members(self, mod):
        mod_name = mod.__name__

        if mod_name == 'dummy':
            mod_name = None

        for k, m in inspect.getmembers(mod):
            if mod_name:
                key = "%s.%s" % (mod_name, k)
            else:
                key = k

            is_class = inspect.isclass(m)
            is_def = self.is_defined_in_module(mod, mod_name, m)

            if not is_def:
                # this is something imported, not defined in the module
                # so we don't want to document it here
                self.log_debug("skipping %s for module %s" % (k, mod_name))
                continue

            if not is_class:
                self.append_item_content(key, m)

            else:
                self.append_item_content(key, m)
                for ck, cm in inspect.getmembers(m):
                    self.append_item_content("%s.%s" % (key, ck), cm)

    def process_module(self, package_name, name):
        self.log_debug("processing module %s" % name)
        mod = self.load_module(name)
        self.append_item_content(name, mod)
        if mod:
            self.process_members(mod)
        else:
            self.log_warn("no mod from %s" % name)

    def process_package(self, package):
        """
        Iterates over all modules included in the package and processes them.
        """
        self.log_debug("processing package %s" % package)
        package_name = package.__name__

        # Process top level package
        self.process_module(package_name, package_name)

        # Process sub-packages and modules
        if hasattr(package, '__path__'):
            path = package.__path__
            prefix = "%s." % package_name
            for loader, name, ispkg in pkgutil.walk_packages(path, prefix=prefix):
                self.process_module(package_name, name)

    def process_packages(self):
        package_names = str(self.input_data).split()
        packages = [__import__(name) for name in package_names]

        for package in packages:
            self.process_package(package)

    def process_file(self):
        mod = self.load_source_file()
        if mod:
            self.process_members(mod)

    def process(self):
        if self.prev_ext == '.txt':
            self.process_packages()
        elif self.prev_ext == '.py':
            self.process_file()
        else:
            raise InternalDexyProblem("Should not have ext %s" % self.prev_ext)

        self.output_data.save()



from tests.utils import wrap
from tests.utils import TEST_DATA_DIR
from dexy.doc import Doc
import os

markdown_file = os.path.join(TEST_DATA_DIR, "markdown-with-python.md")

def run_filter(alias):
    with open(markdown_file, 'r') as f:
        example_markdown = f.read()

    with wrap() as wrapper:
        node = Doc("example.md|%s" % alias,
                wrapper,
                [],
                contents = example_markdown
                )
        wrapper.run_docs(node)
        assert node.output_data().is_cached()
        return node.output_data()


def test_markdown_to_ipynb():
    print(run_filter("mdjup"))

def test_markdown_to_generic():
    print(run_filter("mdsections"))



from tests.utils import wrap
from dexy.doc import Doc

def test_split_html_filter():
    with wrap() as wrapper:
        contents="""
        <p>This is at the top.</p>
        <!-- split "index" -->
        index page content only
        <!-- split "a-page" -->
        some content on a page
        <!-- split "another-page" -->
        This is information about "another-page" which should appear on the index page.
        <!-- content -->
        some content on another page
        <!-- footer -->
        footer on index page only
        <!-- endsplit -->
        bottom
        """

        node = Doc("subdir/example.html|splithtml", wrapper, [], contents=contents)
        wrapper.run_docs(node)

        assert node.children[0].key == "subdir/a-page.html"
        assert node.children[1].key == "subdir/another-page.html"

        od = str(node.output_data())

        assert "<p>This is at the top.</p>" in od
        assert 'index page content only' in od
        assert '<a href="a-page.html">' in od
        assert '<a href="another-page.html">' in od
        assert "This is information about \"another-page\"" in od
        assert "bottom" in od

        od = str(node.children[0].output_data())
        assert "<p>This is at the top.</p>" in od
        assert not 'index page content only' in od
        assert "some content on a page" in od
        assert "bottom" in od

        od = str(node.children[1].output_data())
        assert "<p>This is at the top.</p>" in od
        assert "some content on another page" in od
        assert "bottom" in od

def test_split_html_additional_filters():
    with wrap() as wrapper:
        contents="""
        <p>This is at the top.</p>
        <!-- split "a-page" -->
        some content on a page
        <!-- split "another-page" -->
        some content on another page
        <!-- endsplit -->
        bottom
        """

        node = Doc("example.html|splithtml",
                wrapper,
                [],
                contents=contents,
                splithtml = { "keep-originals" : False, "additional-doc-filters" : "processtext" },
              )
        wrapper.run_docs(node)

        assert node.children[0].key == "a-page.html|processtext"
        assert node.children[1].key == "another-page.html|processtext"

        od = str(node.output_data())
        assert "<p>This is at the top.</p>" in od
        assert '<a href="a-page.html">' in od
        assert '<a href="another-page.html">' in od
        assert "bottom" in od

        a_page = node.children[0]
        a_page_data = str(a_page.output_data())
        assert "<p>This is at the top.</p>" in a_page_data
        assert "some content on a page" in a_page_data
        assert "bottom" in a_page_data
        assert "Dexy processed the text" in a_page_data

        another_page = node.children[1]
        another_page_data = str(another_page.output_data())
        assert "<p>This is at the top.</p>" in another_page_data
        assert "some content on another page" in another_page_data
        assert "bottom" in another_page_data
        assert "Dexy processed the text" in another_page_data



from cashew.exceptions import InactivePlugin
from cashew.exceptions import UserFeedback
from dexy.version import DEXY_VERSION
import dexy.utils
import platform

class NoFilterOutput(UserFeedback):
    pass

class CircularDependency(UserFeedback):
    pass

class BlankAlias(UserFeedback):
    pass

class InvalidStateTransition(Exception):
    pass

class UnexpectedState(Exception):
    pass

class InternalDexyProblem(Exception):
    def __init__(self, message):
        self.message = dexy.utils.s("""
        Oops! You may have found a bug in Dexy.
        The developer would really appreciate if you copy and paste this entire message
        and the Traceback above it into an email and send to info@dexy.it
        Your version of Dexy is %s
        Your platform is %s""" % (DEXY_VERSION, platform.system()))
        self.message += "\n"
        self.message += message

    def __str__(self):
        return self.message

class DeprecatedException(InternalDexyProblem):
    pass

class TemplateException(InternalDexyProblem):
    pass



from tests.utils import tempdir
from dexy.wrapper import Wrapper
import dexy.batch
import os

def test_batch():
    with tempdir():
        wrapper = Wrapper()
        wrapper.create_dexy_dirs()

        wrapper = Wrapper()
        batch = dexy.batch.Batch(wrapper)
        os.makedirs(batch.batch_dir())

        batch.save_to_file()
        assert batch.filename() in os.listdir(".dexy/batches")

        wrapper = Wrapper()
        batch = dexy.batch.Batch.load_most_recent(wrapper)

def test_batch_with_docs():
    with tempdir():
        wrapper = Wrapper(log_level='DEBUG', debug=True)
        wrapper.create_dexy_dirs()

        with open("hello.txt", "w") as f:
            f.write("hello")

        with open("dexy.yaml", "w") as f:
            f.write("hello.txt")

        wrapper = Wrapper()
        wrapper.run_from_new()

        batch = dexy.batch.Batch.load_most_recent(wrapper)
        assert batch

        for doc_key in batch.docs:
            assert batch.input_data(doc_key)
            assert batch.output_data(doc_key)



from dexy.doc import Doc
from tests.utils import TEST_DATA_DIR
from tests.utils import wrap
from tests.utils import capture_stdout
from mock import patch
import dexy.exceptions
import json
import os
import shutil
import dexy.filter
import dexy.wrapper

def test_docmd_create_keyfile():
    with wrap():
        assert not os.path.exists(".dexyapis")
        dexy.filter.Filter.create_instance("wp").docmd_create_keyfile()
        assert os.path.exists(".dexyapis")

def test_docmd_create_keyfile_if_exists():
    with wrap():
        with open(".dexyapis", "w") as f:
            f.write("{}")
        assert os.path.exists(".dexyapis")
        try:
            dexy.filter.Filter.create_instance("wp").docmd_create_keyfile()
            assert False, ' should raise exception'
        except dexy.exceptions.UserFeedback as e:
            assert ".dexyapis already exists" in e.message

def test_api_url_with_php_ending():
    with wrap():
        with open(".dexyapis", "w") as f:
            json.dump({
                    "wordpress" : {"url" : "http://example.com/api/xmlrpc.php"}
                    }, f)

        url = dexy.filter.Filter.create_instance("wp").api_url()
        assert url == "http://example.com/api/xmlrpc.php"

def test_api_url_without_php_ending():
    with wrap():
        with open(".dexyapis", "w") as f:
            json.dump({ "wordpress" : {"url" : "http://example.com/api"} }, f)

        url = dexy.filter.Filter.create_instance("wp").api_url()
        assert url == "http://example.com/api/xmlrpc.php"

def test_api_url_without_php_ending_with_trailing_slash():
    with wrap():
        with open(".dexyapis", "w") as f:
            json.dump({ "wordpress" : {"url" : "http://example.com/api/"} }, f)

        url = dexy.filter.Filter.create_instance("wp").api_url()
        assert url == "http://example.com/api/xmlrpc.php"

def test_wordpress_without_doc_config_file():
    with wrap() as wrapper:
        wrapper.debug = False
        doc = Doc("hello.txt|wp",
                contents = "hello, this is a blog post",
                wrapper=wrapper
                )

        wrapper.run_docs(doc)
        assert wrapper.state == 'error'

def mk_wp_doc(wrapper):
    doc = Doc("hello.txt|wp",
            contents = "hello, this is a blog post",
            dirty = True,
            wrapper=wrapper
            )
    for d in doc.datas():
        d.setup()
    return doc

ATTRS = {
        'return_value.metaWeblog.newPost.return_value' : 42,
        'return_value.metaWeblog.getPost.return_value' : {
            'permaLink' : 'http://example.com/blog/42'
            },
        'return_value.wp.getCategories.return_value' : [
            { 'categoryName' : 'foo' },
            { 'categoryName' : 'bar' }
            ],
        'return_value.wp.uploadFile.return_value' : {
            'url' : 'http://example.com/example.pdf'
            }
        }

@patch('xmlrpc.client.ServerProxy', **ATTRS)
def test_wordpress(MockXmlrpclib):
    with wrap():
        with open("wordpress.json", "w") as f:
            json.dump({}, f)

        with open(".dexyapis", "w") as f:
            json.dump({
                'wordpress' : {
                    'url' : 'http://example.com',
                    'username' : 'foo',
                    'password' : 'bar'
                    }}, f)

        # Create new (unpublished) draft
        wrapper = dexy.wrapper.Wrapper()
        doc = mk_wp_doc(wrapper)
        wrapper.run_docs(doc)

        with open("wordpress.json", "r") as f:
            result = json.load(f)

        assert result['postid'] == 42
        assert result['publish'] == False

        # Update existing draft
        wrapper = dexy.wrapper.Wrapper()
        doc = mk_wp_doc(wrapper)
        wrapper.run_docs(doc)
        assert list(doc.output_data().json_as_dict().keys()) == ['permaLink']

        result['publish'] = True
        with open("wordpress.json", "w") as f:
            json.dump(result, f)

        # Publish existing draft
        wrapper = dexy.wrapper.Wrapper()
        doc = mk_wp_doc(wrapper)
        wrapper.run_docs(doc)
        assert "http://example.com/blog/42" in str(doc.output_data())

        # Now, separately, test an image upload.
        orig = os.path.join(TEST_DATA_DIR, 'color-graph.pdf')
        shutil.copyfile(orig, 'example.pdf')
        from dexy.wrapper import Wrapper
        wrapper = Wrapper()
        doc = Doc("example.pdf|wp",
                wrapper=wrapper)

        with open(".dexyapis", "w") as f:
            json.dump({
                'wordpress' : {
                    'url' : 'http://example.com',
                    'username' : 'foo',
                    'password' : 'bar'
                    }}, f)

        wrapper.run_docs(doc)
        assert doc.output_data().as_text() == "http://example.com/example.pdf"

        # test list categories
        with capture_stdout() as stdout:
            dexy.filter.Filter.create_instance("wp").docmd_list_categories()
            assert stdout.getvalue() == "categoryName\nfoo\nbar\n"



from dexy.filter import DexyFilter
from dexy.utils import indent
import copy
import dexy.exceptions
import json
import os
import re
import textwrap

class Resub(DexyFilter):
    """
    Runs re.sub on each line of input.
    """
    aliases = ['resub']
    _settings = {
            'expressions' : ("Tuples of (regexp, replacement) to apply.", []),
            }

    def process_text(self, input_text):
        for regexp, replacement in self.setting('expressions'):
            self.log_debug("Applying %s" % regexp)
            working_text = []

            for line in input_text.splitlines():
                working_text.append(re.sub(regexp, replacement, line))

            input_text = "\n".join(working_text)

        return input_text


class PreserveDataClassFilter(DexyFilter):
    """
    Sets PRESERVE_PRIOR_DATA_CLASS to True.
    """
    aliases = []
    _settings = {
            'preserve-prior-data-class' : True
            }

    def data_class_alias(self, ext):
        if self.setting('preserve-prior-data-class'):
            return self.input_data.alias
        else:
            return self.setting('data-type')

    def calculate_canonical_name(self):
        return self.prev_filter.calculate_canonical_name()

class ChangeExtensionManuallyFilter(PreserveDataClassFilter):
    """
    Dummy filter for allowing changing a file extension.
    """
    aliases = ['chext']

class KeyValueStoreFilter(DexyFilter):
    """
    Creates a new key-value store.

    The key-value store will be populated via side effects from other filters.
    """
    aliases = ['kv']
    _settings = {
            'data-type' : 'keyvalue'
            }

    def process(self):
        self.output_data.copy_from_file(self.input_data.storage.data_file())

        # Call setup() again since it will have created a new blank database.
        self.output_data.storage.setup()
        self.output_data.storage.connect()

class HeaderFilter(DexyFilter):
    """
    Apply another file to top of file.
    """
    aliases = ['hd']
    _settings = {
            'key-name' : ("Name of key to use.", 'header'),
            'header' : ("Document key of file to use as header.", None)
            }

    def find_input_in_parent_dir(self, matches):
        docs = list(self.doc.walk_input_docs())
        docs_d = dict((task.output_data().long_name(), task) for task in docs)

        key_name = self.setting('key-name')
        requested = self.setting(key_name)
        if requested:
            if requested in docs_d:
                matched_key = requested
            else:
                msg = "Couldn't find the %s file %s you requested" % (self.setting(key_name), requested)
                raise dexy.exceptions.UserFeedback(msg)
        else:
            matched_key = None
            for k in sorted(docs_d.keys()):
                if (os.path.dirname(k) in self.output_data.parent_dir()) and (matches in k):
                    matched_key = k

        if not matched_key:
            msg = "no %s input found for %s" 
            msgargs = (self.setting('key-name'), self.key)
            raise dexy.exceptions.UserFeedback(msg % msgargs)

        return docs_d[matched_key].output_data()

    def process_text(self, input_text):
        header_data = self.find_input_in_parent_dir("_header")
        return "%s\n%s" % (str(header_data), input_text)

class FooterFilter(HeaderFilter):
    """
    Apply another file to bottom of file.
    """
    aliases = ['ft']
    _settings = {
            'key-name' : 'footer',
            'footer' : ("Document key of file to use as footer.", None)
            }

    def process_text(self, input_text):
        footer_data = self.find_input_in_parent_dir("_footer")
        return "%s\n%s" % (input_text, str(footer_data))

class TemplateContentFilter(HeaderFilter):
    """
    Apply template to file. Template should specify %(content)s.
    """
    aliases = ['applytemplate']
    _settings = {
            'key-name' : 'template',
            'template' : ("Document key of file to use as template.", None)
            }

    def process_text(self, input_text):
        template_data = self.find_input_in_parent_dir("_template")
        return str(template_data) % { 'content' : input_text, 'title' : self.input_data.title()}

class FormatTemplateContentFilter(HeaderFilter):
    """
    Apply template to file. Template should specify {content}.
    """
    aliases = ['formattemplate']
    _settings = {
            'key-name' : 'template',
            'template' : ("Document key of file to use as footer.", None)
            }

    def process_text(self, input_text):
        template_data = self.find_input_in_parent_dir("_template")
        return str(template_data).format({ 'content' : input_text, 'title' : self.input_data.title()})

class MarkupTagsFilter(DexyFilter):
    """
    Wrap text in specified HTML tags.
    """
    aliases = ['tags']
    _settings = {
            'tags' : ("Tags.", {})
            }

    def process_text(self, input_text):
        tags = copy.copy(self.setting('tags'))
        open_tags = "".join("<%s>" % t for t in tags)
        tags.reverse()
        close_tags = "".join("</%s>" % t for t in tags)

        return "%s\n%s\n%s" % (open_tags, input_text, close_tags)

class StartSpaceFilter(DexyFilter):
    """
    Add a blank space to the start of each line.

    Useful for passing syntax highlighted/preformatted code to mediawiki.
    """
    aliases = ['ss', 'startspace']
    _settings = {
            'n' : ("Number of spaces to prepend to each line.", 1),
            'data-type' : 'sectioned'
            }

    def process(self):
        n = self.setting('n')
        for section_name, section_input in self.input_data.items():
            self.output_data[section_name] = indent(section_input, n)
        self.output_data.save()

class SectionsByLine(DexyFilter):
    """
    Returns each line in its own section.
    """
    aliases = ['lines']
    _settings = {
            'data-type' : 'sectioned'
            }

    def process(self):
        input_text = str(self.input_data)
        for i, line in enumerate(input_text.splitlines()):
            self.output_data["%s" % (i+1)] = line
        self.output_data.save()

class ClojureWhitespaceFilter(DexyFilter):
    """
    Parse clojure code into sections based on whitespace and try to guess a
    useful name for each section by looking for def, defn, deftest or a
    comment.
    """
    aliases = ['cljws']
    _settings = {
            'added-in-version' : '1.0.1',
            'data-type' : 'sectioned',
            'name-regex' : (
                """List of regular expressions including a match group
                representing the name of the section. Will be tried in
                order.""",
                [
                    "\(def ([a-z\-\?]+)",
                    "\(defn ([a-z\-\?]+)",
                    "\(defn- ([a-z\-\?]+)",
                    "\(defn= ([a-z\-\?]+)",
                    "\(deftest ([a-z\-\?]+)",
                ])
            }

    def process(self):
        input_text = str(self.input_data)
        for i, section in enumerate(input_text.split("\n\n")):
            section_name = self.parse_section_name(section)
            if section_name:
                self.output_data[section_name] = section
            else:
                self.output_data["%s" % (i+1)] = section

        self.output_data.save()

    def parse_section_name(self, section_text):
        """
        Parse a section name out of the section text.
        """
        if not section_text:
            return

        for line in section_text.splitlines():
            firstline = line
            if not line.strip().startswith(';'):
                break

        for regex in self.setting('name-regex'):
            m = re.match(regex, firstline)
            if m:
                return m.groups()[0]

class PrettyPrintJsonFilter(DexyFilter):
    """
    Pretty prints JSON input.
    """
    aliases = ['ppjson']
    _settings = {
            'output-extensions' : ['.json']
            }

    def process_text(self, input_text):
        json_content = json.loads(input_text)
        return json.dumps(json_content, sort_keys=True, indent=4)

class JoinFilter(DexyFilter):
    """
    Takes sectioned code and joins it into a single section. Some filters which
    don't preserve sections will raise an error if they receive multiple
    sections as input, so this forces acknowledgement that sections will be
    lost.
    """
    aliases = ['join']

    def process(self):
        joined_data = "\n".join(str(v) for v in list(self.input_data.values()))
        print("joined data is", joined_data)
        self.output_data.set_data(joined_data)

class HeadFilter(DexyFilter):
    """
    Returns just the first 10 lines of input.
    """
    aliases = ['head']

    def process_text(self, input_text):
        return "\n".join(input_text.split("\n")[0:10]) + "\n"

class WordWrapFilter(DexyFilter):
    """
    Wraps text after 79 characters (tries to preserve existing line breaks and
    spaces).
    """
    aliases = ['ww', 'wrap']
    _settings = {
            'width' : ("Width of text to wrap to.", 79)
            }

    def wrap_text(self, text, width):
        """
        A word-wrap function that preserves existing line breaks
        and most spaces in the text. Expects that existing line
        breaks are posix newlines (\n).
        """
        return "\n".join(textwrap.wrap(text, width))

    def process_text(self, input_text):
        return self.wrap_text(input_text, self.setting('width'))



from tests.utils import runfilter
from dexy.exceptions import UserFeedback

min_valid_html = """
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2 Final//EN">
<TITLE></TITLE>
"""

invalid_html = "<html></html>"

def test_tidyerrors():
    with runfilter("tidyerrors", invalid_html, ext=".html") as doc:
        output = str(doc.output_data())
        assert "missing <!DOCTYPE>" in output
        assert "inserting missing 'title'" in output

def test_htmltidy_throws_error_on_invalid_html():
    try:
        with runfilter("htmltidy", invalid_html, ext=".html"):
            assert False, "should not get here"
    except UserFeedback as e:
        assert "missing <!DOCTYPE>" in str(e)
        assert "inserting missing 'title'" in str(e)

def test_htmltidy():
    with runfilter("htmltidy", min_valid_html, ext=".html") as doc:
        output = str(doc.output_data())
        assert "<body>" in output

def test_htmlcheck_on_valid_html():
    with runfilter("tidycheck", min_valid_html, ext=".html") as doc:
        output = str(doc.output_data())
        assert output == min_valid_html

def test_htmlcheck_on_invalid_html():
    try:
        with runfilter("tidycheck", invalid_html, ext=".html"):
            assert False, "should not get here"
    except UserFeedback as e:
        assert "missing <!DOCTYPE>" in str(e)
        assert "inserting missing 'title'" in str(e)



from tests.utils import wrap
from dexy.doc import Doc

HTML = """
<h1>Header</h1>
<p>This is some body.</p>
"""

def run_calibre(ext):
    with wrap() as wrapper:
        node = Doc("book.html|calibre",
                wrapper,
                [],
                calibre = { 'ext' : ext },
                contents = HTML
                )
        wrapper.run_docs(node)
        assert node.output_data().is_cached()

def test_calibre_mobi():
    run_calibre('.mobi')

def test_calibre_epub():
    run_calibre('.epub')

def test_calibre_fb2():
    run_calibre('.fb2')

def test_calibre_htmlz():
    run_calibre('.htmlz')

def test_calibre_lit():
    run_calibre('.lit')

def test_calibre_lrf():
    run_calibre('.lrf')

def test_calibre_pdf():
    run_calibre('.pdf')

def test_calibre_rtf():
    run_calibre('.rtf')

def test_calibre_snb():
    run_calibre('.snb')

def test_calibre_tcr():
    run_calibre('.tcr')

def test_calibre_txt():
    run_calibre('.txt')

def test_calibre_txtz():
    run_calibre('.txtz')
